"""Miscallenous helper functionality."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2008 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

import contextlib

import numpy as np

import pycuda._driver as _drv
import pycuda.driver as cuda
from pycuda.compyte.dtypes import (  # noqa: F401
    _fill_dtype_registry,
    dtype_to_ctype as base_dtype_to_ctype,
    get_or_register_dtype,
    register_dtype,
)


bitlog2 = _drv.bitlog2
DeviceMemoryPool = _drv.DeviceMemoryPool
PageLockedMemoryPool = _drv.PageLockedMemoryPool
PageLockedAllocator = _drv.PageLockedAllocator

_fill_dtype_registry(respect_windows=True)
get_or_register_dtype("pycuda::complex<float>", np.complex64)
get_or_register_dtype("pycuda::complex<double>", np.complex128)


# {{{ debug memory pool


class DebugMemoryPool(DeviceMemoryPool):
    def __init__(self, interactive=True, logfile=None):
        DeviceMemoryPool.__init__(self)
        self.last_free, _ = cuda.mem_get_info()
        self.interactive = interactive

        if logfile is None:
            import sys

            logfile = sys.stdout

        self.logfile = logfile

        from weakref import WeakKeyDictionary

        self.blocks = WeakKeyDictionary()

        if interactive:
            from pytools.diskdict import DiskDict

            self.stacktrace_mnemonics = DiskDict("pycuda-stacktrace-mnemonics")

    def allocate(self, size):
        from traceback import extract_stack

        stack = tuple(frm[2] for frm in extract_stack())
        description = self.describe(stack, size)

        histogram = {}
        for bsize, descr in self.blocks.values():
            histogram[bsize, descr] = histogram.get((bsize, descr), 0) + 1

        from pytools import common_prefix

        cpfx = common_prefix(descr for bsize, descr in histogram)

        print(
            "\n  Allocation of size %d occurring "
            "(mem: last_free:%d, free: %d, total:%d) (pool: held:%d, active:%d):"
            "\n      at: %s"
            % (
                (size, self.last_free, *cuda.mem_get_info(),
                    self.held_blocks, self.active_blocks, description)
            ),
            file=self.logfile,
        )

        hist_items = sorted(histogram.items())
        for (bsize, descr), count in hist_items:
            print(
                "  %s (%d bytes): %dx" % (descr[len(cpfx):], bsize, count),
                file=self.logfile,
            )

        if self.interactive:
            input("  [Enter]")

        result = DeviceMemoryPool.allocate(self, size)
        self.blocks[result] = size, description
        self.last_free, _ = cuda.mem_get_info()
        return result

    def describe(self, stack, size):
        if not self.interactive:
            return "|".join(stack)
        else:
            try:
                return self.stacktrace_mnemonics[stack, size]
            except KeyError:
                print(size, stack)
                while True:
                    mnemonic = input("Enter mnemonic or [Enter] for more info:")
                    if mnemonic == "":
                        from traceback import print_stack

                        print_stack()
                    else:
                        break
                self.stacktrace_mnemonics[stack, size] = mnemonic
                return mnemonic


# }}}


# {{{ default device/context


def get_default_device(default=0):
    from warnings import warn

    warn(
        "get_default_device() is deprecated; " "use make_default_context() instead",
        DeprecationWarning, stacklevel=2,
    )

    import os

    from pycuda.driver import Device

    dev = os.environ.get("CUDA_DEVICE")

    if dev is None:
        with contextlib.suppress(Exception):  # noqa: SIM117
            with open(os.path.join(os.path.expanduser("~"), ".cuda_device")) as devrc:
                dev = devrc.read().strip()

    if dev is None:
        dev = default

    try:
        dev = int(dev)
    except TypeError:
        raise TypeError(
            "CUDA device number (CUDA_DEVICE or ~/.cuda-device) " "must be an integer"
        )

    return Device(dev)


def make_default_context(ctx_maker=None):
    if ctx_maker is None:

        def ctx_maker(dev):
            return dev.make_context()

    ndevices = cuda.Device.count()
    if ndevices == 0:
        raise RuntimeError(
            "No CUDA enabled device found. " "Please check your installation."
        )

    # Is CUDA_DEVICE set?
    import os

    devn = os.environ.get("CUDA_DEVICE")

    # Is $HOME/.cuda_device set ?
    if devn is None:
        try:
            homedir = os.environ.get("HOME")
            assert homedir is not None
            with open(os.path.join(homedir, ".cuda_device")) as devrc:
                devn = devrc.read().strip()
        except Exception:
            pass

    # If either CUDA_DEVICE or $HOME/.cuda_device is set, try to use it
    if devn is not None:
        try:
            devn = int(devn)
        except TypeError:
            raise TypeError(
                "CUDA device number (CUDA_DEVICE or ~/.cuda_device)"
                " must be an integer"
            )

        dev = cuda.Device(devn)
        return ctx_maker(dev)

    # Otherwise, try to use any available device
    else:
        for devn in range(ndevices):
            dev = cuda.Device(devn)
            try:
                return ctx_maker(dev)
            except cuda.Error:
                pass

        raise RuntimeError(
            "make_default_context() wasn't able to create a context "
            "on any of the %d detected devices" % ndevices
        )


# }}}


# {{{ rounding helpers


def _exact_div(dividend, divisor):
    quot, rem = divmod(dividend, divisor)
    assert rem == 0
    return quot


def _int_ceiling(value, multiple_of=1):
    """Round C{value} up to be a C{multiple_of} something."""
    # Mimicks the Excel "floor" function (for code stolen from occupancy calculator)

    from math import ceil

    return ceil(value / multiple_of) * multiple_of


def _int_floor(value, multiple_of=1):
    """Round C{value} down to be a C{multiple_of} something."""
    # Mimicks the Excel "floor" function (for code stolen from occupancy calculator)

    from math import floor

    return floor(value / multiple_of) * multiple_of


# }}}


# {{{ device data


class DeviceData:
    def __init__(self, dev=None):
        import pycuda.driver as drv

        if dev is None:
            dev = cuda.Context.get_device()

        self.max_threads = dev.get_attribute(drv.device_attribute.MAX_THREADS_PER_BLOCK)
        self.warp_size = dev.get_attribute(drv.device_attribute.WARP_SIZE)

        if dev.compute_capability() >= (3, 0):
            self.warps_per_mp = 64
        elif dev.compute_capability() >= (2, 0):
            self.warps_per_mp = 48
        elif dev.compute_capability() >= (1, 2):
            self.warps_per_mp = 32
        else:
            self.warps_per_mp = 24

        self.thread_blocks_per_mp = 8
        self.registers = dev.get_attribute(drv.device_attribute.MAX_REGISTERS_PER_BLOCK)
        self.shared_memory = dev.get_attribute(
            drv.device_attribute.MAX_SHARED_MEMORY_PER_BLOCK
        )

        if dev.compute_capability() >= (2, 0):
            self.smem_alloc_granularity = 128
            self.smem_granularity = 32
        else:
            self.smem_alloc_granularity = 512
            self.smem_granularity = 16

        if dev.compute_capability() >= (2, 0):
            self.register_allocation_unit = "warp"
        else:
            self.register_allocation_unit = "block"

    def align(self, bytes, word_size=4):
        return _int_ceiling(bytes, self.align_bytes(word_size))

    def align_dtype(self, elements, dtype_size):
        return _int_ceiling(elements, self.align_words(dtype_size))

    def align_words(self, word_size):
        return _exact_div(self.align_bytes(word_size), word_size)

    def align_bytes(self, word_size=4):
        if word_size == 4:
            return 64
        elif word_size == 8 or word_size == 16:
            return 128
        else:
            raise ValueError("no alignment possible for fetches of size %d" % word_size)

    def coalesce(self, thread_count):
        return _int_ceiling(thread_count, 16)

    @staticmethod
    def make_valid_tex_channel_count(size):
        valid_sizes = [1, 2, 4]
        for vs in valid_sizes:
            if size <= vs:
                return vs

        raise ValueError("could not enlarge argument to valid channel count")


# }}}

# {{{ occupancy


class OccupancyRecord:
    def __init__(self, devdata, threads, shared_mem=0, registers=0):
        if threads > devdata.max_threads:
            raise ValueError("too many threads")

        # copied literally from occupancy calculator
        alloc_warps = _int_ceiling(threads / devdata.warp_size)
        alloc_smem = _int_ceiling(shared_mem, devdata.smem_alloc_granularity)
        if devdata.register_allocation_unit == "warp":
            alloc_regs = alloc_warps * 32 * registers
        elif devdata.register_allocation_unit == "block":
            alloc_regs = _int_ceiling(alloc_warps * 2, 4) * 16 * registers
        else:
            raise ValueError(
                "Improper register allocation unit:" + devdata.register_allocation_unit
            )

        if alloc_regs > devdata.registers:
            raise ValueError("too many registers")

        if alloc_smem > devdata.shared_memory:
            raise ValueError("too much smem")

        self.tb_per_mp_limits = [
            (devdata.thread_blocks_per_mp, "device"),
            (_int_floor(devdata.warps_per_mp / alloc_warps), "warps"),
        ]
        if registers > 0:
            self.tb_per_mp_limits.append(
                (_int_floor(devdata.registers / alloc_regs), "regs")
            )
        if shared_mem > 0:
            self.tb_per_mp_limits.append(
                (_int_floor(devdata.shared_memory / alloc_smem), "smem")
            )

        self.tb_per_mp, self.limited_by = min(self.tb_per_mp_limits)

        self.warps_per_mp = self.tb_per_mp * alloc_warps
        self.occupancy = self.warps_per_mp / devdata.warps_per_mp


# }}}

# {{{ C types <-> dtypes


class Argument:
    def __init__(self, dtype, name):
        self.dtype = np.dtype(dtype)
        self.name = name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r}, {self.dtype})"


def dtype_to_ctype(dtype, with_fp_tex_hack=False):
    if dtype is None:
        raise ValueError("dtype may not be None")

    dtype = np.dtype(dtype)
    if with_fp_tex_hack:
        if dtype == np.float32:
            return "fp_tex_float"
        elif dtype == np.float64:
            return "fp_tex_double"
        elif dtype == np.complex64:
            return "fp_tex_cfloat"
        elif dtype == np.complex128:
            return "fp_tex_cdouble"

    return base_dtype_to_ctype(dtype)


class VectorArg(Argument):
    def declarator(self):
        return "{} *{}".format(dtype_to_ctype(self.dtype), self.name)

    struct_char = "P"


class ScalarArg(Argument):
    def declarator(self):
        return "{} {}".format(dtype_to_ctype(self.dtype), self.name)

    @property
    def struct_char(self):
        result = self.dtype.char
        if result == "V":
            result = "%ds" % self.dtype.itemsize

        return result


def parse_c_arg(c_arg):
    from pycuda.compyte.dtypes import parse_c_arg_backend

    return parse_c_arg_backend(c_arg, ScalarArg, VectorArg)


def get_arg_type(c_arg):
    return parse_c_arg(c_arg).struct_char


# }}}

# {{{ context-dep memoization

context_dependent_memoized_functions = []


def context_dependent_memoize(func):
    def wrapper(*args, **kwargs):
        cache_key = (args, frozenset(kwargs.items())) if kwargs else (args,)

        try:
            ctx_dict = func._pycuda_ctx_dep_memoize_dic
        except AttributeError:
            # FIXME: This may keep contexts alive longer than desired.
            # But I guess since the memory in them is freed, who cares.
            ctx_dict = func._pycuda_ctx_dep_memoize_dic = {}

        cur_ctx = cuda.Context.get_current()

        try:
            return ctx_dict[cur_ctx][cache_key]
        except KeyError:
            context_dependent_memoized_functions.append(func)
            arg_dict = ctx_dict.setdefault(cur_ctx, {})
            result = func(*args, **kwargs)
            arg_dict[cache_key] = result
            return result

    from functools import update_wrapper
    update_wrapper(wrapper, func)
    return wrapper


def clear_context_caches():
    for func in context_dependent_memoized_functions:
        try:
            ctx_dict = func._pycuda_ctx_dep_memoize_dic
        except AttributeError:
            pass
        else:
            ctx_dict.clear()


# }}}

# {{{ py.test interaction


def mark_cuda_test(inner_f):
    def f(*args, **kwargs):
        import pycuda.driver

        # appears to be idempotent, i.e. no harm in calling it more than once
        pycuda.driver.init()

        ctx = make_default_context()
        try:
            assert isinstance(ctx.get_device().name(), str)
            assert isinstance(ctx.get_device().compute_capability(), tuple)
            assert isinstance(ctx.get_device().get_attributes(), dict)
            inner_f(*args, **kwargs)
        finally:
            ctx.pop()

            from pycuda.tools import clear_context_caches

            clear_context_caches()

            from gc import collect

            collect()

    try:
        from py.test import mark as mark_test
    except ImportError:
        return f

    return mark_test.cuda(f)


def init_cuda_context_fixture():
    import pycuda.driver as cuda
    cuda.init()
    ctx = make_default_context()
    assert isinstance(ctx.get_device().name(), str)
    assert isinstance(ctx.get_device().compute_capability(), tuple)
    assert isinstance(ctx.get_device().get_attributes(), dict)
    yield

    from gc import collect
    ctx.pop()
    clear_context_caches()
    collect()

# }}}


# vim: foldmethod=marker



from __future__ import annotations

import atexit

import pycuda.driver as cuda
import pycuda.gl as cudagl


cuda.init()
assert cuda.Device.count() >= 1

from pycuda.tools import make_default_context


context = make_default_context(lambda dev: cudagl.make_context(dev))
device = context.get_device()

atexit.register(context.pop)



"""Elementwise functionality."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2009 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""


from typing import Any

import numpy as np

from pytools import memoize_method

from pycuda.tools import ScalarArg, VectorArg, context_dependent_memoize, dtype_to_ctype


def get_elwise_module(
    arguments,
    operation,
    name="kernel",
    keep=False,
    options=None,
    preamble="",
    loop_prep="",
    after_loop="",
):
    from pycuda.compiler import SourceModule
    return SourceModule(
        """
        #include <pycuda-complex.hpp>

        %(preamble)s

        extern "C"
        __global__ void %(name)s(%(arguments)s)
        {

          size_t tid = threadIdx.x;
          size_t total_threads = gridDim.x*blockDim.x;
          size_t cta_start = blockDim.x*blockIdx.x;
          size_t i;

          %(loop_prep)s;

          for (i = cta_start + tid; i < n; i += total_threads)
          {
            %(operation)s;
          }

          %(after_loop)s;
        }
        """
        % {
            "arguments": ", ".join(arg.declarator() for arg in arguments),
            "operation": operation,
            "name": name,
            "preamble": preamble,
            "loop_prep": loop_prep,
            "after_loop": after_loop,
        },
        options=options,
        keep=keep,
        no_extern_c=True,
    )


def get_elwise_range_module(
    arguments,
    operation,
    name="kernel",
    keep=False,
    options=None,
    preamble="",
    loop_prep="",
    after_loop="",
):
    from pycuda.compiler import SourceModule

    return SourceModule(
        """
        #include <pycuda-complex.hpp>

        %(preamble)s

        extern "C"
        __global__ void %(name)s(%(arguments)s)
        {
          size_t tid = threadIdx.x;
          size_t total_threads = gridDim.x*blockDim.x;
          size_t cta_start = blockDim.x*blockIdx.x;
          long i;

          %(loop_prep)s;

          if (step < 0)
          {
            for (i = start + (cta_start + tid)*step;
              i > stop; i += total_threads*step)
            {
              %(operation)s;
            }
          }
          else
          {
            for (i = start + (cta_start + tid)*step;
              i < stop; i += total_threads*step)
            {
              %(operation)s;
            }
          }

          %(after_loop)s;
        }
        """
        % {
            "arguments": ", ".join(arg.declarator() for arg in arguments),
            "operation": operation,
            "name": name,
            "preamble": preamble,
            "loop_prep": loop_prep,
            "after_loop": after_loop,
        },
        options=options,
        keep=keep,
        no_extern_c=True,
    )


def get_elwise_kernel_and_types(
    arguments,
    operation,
    name="kernel",
    keep=False,
    options=None,
    use_range=False,
    **kwargs
):
    if isinstance(arguments, str):
        from pycuda.tools import parse_c_arg

        arguments = [parse_c_arg(arg) for arg in arguments.split(",")]

    if use_range:
        arguments.extend(
            [
                ScalarArg(np.intp, "start"),
                ScalarArg(np.intp, "stop"),
                ScalarArg(np.intp, "step"),
            ]
        )
    else:
        arguments.append(ScalarArg(np.uintp, "n"))

    module_builder = get_elwise_range_module if use_range else get_elwise_module

    mod = module_builder(arguments, operation, name, keep, options, **kwargs)

    func = mod.get_function(name)
    func.prepare("".join(arg.struct_char for arg in arguments))

    return mod, func, arguments


def get_elwise_kernel(
    arguments, operation, name="kernel", keep=False, options=None, **kwargs
):
    """Return a L{pycuda.driver.Function} that performs the same scalar operation
    on one or several vectors.
    """
    _mod, func, arguments = get_elwise_kernel_and_types(
        arguments, operation, name, keep, options, **kwargs
    )

    return func


class ElementwiseKernel:
    def __init__(
        self, arguments, operation, name="kernel", keep=False, options=None, **kwargs
    ):

        self.gen_kwargs = kwargs.copy()
        self.gen_kwargs.update({
            "keep": keep,
            "options": options,
            "name": name,
            "operation": operation,
            "arguments": arguments,
            })

    def get_texref(self, name, use_range=False):
        mod, _knl, _arguments = self.generate_stride_kernel_and_types(use_range=use_range)
        return mod.get_texref(name)

    @memoize_method
    def generate_stride_kernel_and_types(self, use_range):
        mod, knl, arguments = get_elwise_kernel_and_types(
            use_range=use_range, **self.gen_kwargs
        )

        assert [i for i, arg in enumerate(arguments) if isinstance(arg, VectorArg)], (
            "ElementwiseKernel can only be used with functions that "
            "have at least one vector argument"
        )

        return mod, knl, arguments

    def __call__(self, *args, **kwargs):
        vectors = []

        range_ = kwargs.pop("range", None)
        slice_ = kwargs.pop("slice", None)
        stream = kwargs.pop("stream", None)

        if kwargs:
            raise TypeError(
                "invalid keyword arguments specified: "
                + ", ".join(kwargs.keys())
            )

        invocation_args = []
        _mod, func, arguments = self.generate_stride_kernel_and_types(
            range_ is not None or slice_ is not None
        )

        for arg, arg_descr in zip(args, arguments, strict=False):
            if isinstance(arg_descr, VectorArg):
                if not arg.flags.forc:
                    raise RuntimeError(
                        "elementwise kernel cannot " "deal with non-contiguous arrays"
                    )

                vectors.append(arg)
                invocation_args.append(arg.gpudata)
            else:
                invocation_args.append(arg)

        repr_vec = vectors[0]

        if slice_ is not None:
            if range_ is not None:
                raise TypeError(
                    "may not specify both range and slice " "keyword arguments"
                )

            range_ = slice(*slice_.indices(repr_vec.size))

        if range_ is not None:
            invocation_args.append(range_.start)
            invocation_args.append(range_.stop)
            if range_.step is None:
                invocation_args.append(1)
            else:
                invocation_args.append(range_.step)

            from pycuda.gpuarray import splay

            grid, block = splay(abs(range_.stop - range_.start) // range_.step)
        else:
            block = repr_vec._block
            grid = repr_vec._grid
            invocation_args.append(repr_vec.mem_size)

        func.prepared_async_call(grid, block, stream, *invocation_args)


@context_dependent_memoize
def get_take_kernel(dtype, idx_dtype, vec_count=1):
    ctx = {
        "idx_tp": dtype_to_ctype(idx_dtype),
        "tp": dtype_to_ctype(dtype),
        "tex_tp": dtype_to_ctype(dtype, with_fp_tex_hack=True),
    }

    args = (
        [VectorArg(idx_dtype, "idx")]
        + [VectorArg(dtype, "dest" + str(i)) for i in range(vec_count)]
        + [ScalarArg(np.intp, "n")]
    )
    preamble = "#include <pycuda-helpers.hpp>\n\n" + "\n".join(
        "texture <%s, 1, cudaReadModeElementType> tex_src%d;" % (ctx["tex_tp"], i)
        for i in range(vec_count)
    )
    body = ("%(idx_tp)s src_idx = idx[i];\n" % ctx) + "\n".join(
        "dest%d[i] = fp_tex1Dfetch(tex_src%d, src_idx);" % (i, i)
        for i in range(vec_count)
    )

    mod = get_elwise_module(args, body, "take", preamble=preamble)
    func = mod.get_function("take")
    tex_src = [mod.get_texref("tex_src%d" % i) for i in range(vec_count)]
    func.prepare("P" + (vec_count * "P") + np.dtype(np.uintp).char, texrefs=tex_src)
    return func, tex_src


@context_dependent_memoize
def get_take_put_kernel(dtype, idx_dtype, with_offsets, vec_count=1):
    ctx = {
        "idx_tp": dtype_to_ctype(idx_dtype),
        "tp": dtype_to_ctype(dtype),
        "tex_tp": dtype_to_ctype(dtype, with_fp_tex_hack=True),
    }

    args = (
        [
            VectorArg(idx_dtype, "gmem_dest_idx"),
            VectorArg(idx_dtype, "gmem_src_idx"),
        ]
        + [VectorArg(dtype, "dest%d" % i) for i in range(vec_count)]
        + [
            ScalarArg(idx_dtype, "offset%d" % i)
            for i in range(vec_count)
            if with_offsets
        ]
        + [ScalarArg(np.intp, "n")]
    )

    preamble = "#include <pycuda-helpers.hpp>\n\n" + "\n".join(
        "texture <%s, 1, cudaReadModeElementType> tex_src%d;" % (ctx["tex_tp"], i)
        for i in range(vec_count)
    )

    if with_offsets:

        def get_copy_insn(i):
            return (
                "dest%d[dest_idx] = "
                "fp_tex1Dfetch(tex_src%d, src_idx+offset%d);" % (i, i, i)
            )

    else:

        def get_copy_insn(i):
            return "dest%d[dest_idx] = " "fp_tex1Dfetch(tex_src%d, src_idx);" % (i, i)

    body = (
        "%(idx_tp)s src_idx = gmem_src_idx[i];\n"
        "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx
    ) + "\n".join(get_copy_insn(i) for i in range(vec_count))

    mod = get_elwise_module(args, body, "take_put", preamble=preamble)
    func = mod.get_function("take_put")
    tex_src = [mod.get_texref("tex_src%d" % i) for i in range(vec_count)]

    func.prepare(
        "PP"
        + (vec_count * "P")
        + (bool(with_offsets) * vec_count * idx_dtype.char)
        + np.dtype(np.uintp).char,
        texrefs=tex_src,
    )
    return func, tex_src


@context_dependent_memoize
def get_put_kernel(dtype, idx_dtype, vec_count=1):
    ctx = {
        "idx_tp": dtype_to_ctype(idx_dtype),
        "tp": dtype_to_ctype(dtype),
    }

    args = (
        [
            VectorArg(idx_dtype, "gmem_dest_idx"),
        ]
        + [VectorArg(dtype, "dest%d" % i) for i in range(vec_count)]
        + [VectorArg(dtype, "src%d" % i) for i in range(vec_count)]
        + [ScalarArg(np.intp, "n")]
    )

    body = "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx + "\n".join(
        "dest%d[dest_idx] = src%d[i];" % (i, i) for i in range(vec_count)
    )

    func = get_elwise_module(args, body, "put").get_function("put")
    func.prepare("P" + (2 * vec_count * "P") + np.dtype(np.uintp).char)
    return func


@context_dependent_memoize
def get_copy_kernel(dtype_dest, dtype_src):
    return get_elwise_kernel(
        "%(tp_dest)s *dest, %(tp_src)s *src"
        % {
            "tp_dest": dtype_to_ctype(dtype_dest),
            "tp_src": dtype_to_ctype(dtype_src),
        },
        "dest[i] = src[i]",
        "copy",
    )


@context_dependent_memoize
def get_linear_combination_kernel(summand_descriptors, dtype_z):
    from pycuda.elementwise import ScalarArg, VectorArg, get_elwise_module
    from pycuda.tools import dtype_to_ctype

    args = []
    preamble = ["#include <pycuda-helpers.hpp>\n\n"]
    loop_prep = []
    summands = []
    tex_names = []

    for i, (is_gpu_scalar, scalar_dtype, vector_dtype) in enumerate(
        summand_descriptors
    ):
        if is_gpu_scalar:
            preamble.append(
                "texture <%s, 1, cudaReadModeElementType> tex_a%d;"
                % (dtype_to_ctype(scalar_dtype, with_fp_tex_hack=True), i)
            )
            args.append(VectorArg(vector_dtype, "x%d" % i))
            tex_names.append("tex_a%d" % i)
            loop_prep.append(
                "%s a%d = fp_tex1Dfetch(tex_a%d, 0)"
                % (dtype_to_ctype(scalar_dtype), i, i)
            )
        else:
            args.append(ScalarArg(scalar_dtype, "a%d" % i))
            args.append(VectorArg(vector_dtype, "x%d" % i))

        summands.append("a%d*x%d[i]" % (i, i))

    args.append(VectorArg(dtype_z, "z"))
    args.append(ScalarArg(np.uintp, "n"))

    mod = get_elwise_module(
        args,
        "z[i] = " + " + ".join(summands),
        "linear_combination",
        preamble="\n".join(preamble),
        loop_prep=";\n".join(loop_prep),
    )

    func = mod.get_function("linear_combination")
    tex_src = [mod.get_texref(tn) for tn in tex_names]
    func.prepare("".join(arg.struct_char for arg in args), texrefs=tex_src)

    return func, tex_src


def _get_real_dtype(dtype: np.dtype[Any]) -> np.dtype[Any]:
    assert dtype.kind == "c"
    return np.empty(0, dtype).real.dtype


@context_dependent_memoize
def get_axpbyz_kernel(dtype_x, dtype_y, dtype_z,
                    x_is_scalar=False, y_is_scalar=False):
    """
    Returns a kernel corresponding to ``z = ax + by``.

    :arg x_is_scalar: A :class:`bool` which is *True* only if `x` is a scalar :class:`gpuarray`.
    :arg y_is_scalar: A :class:`bool` which is *True* only if `y` is a scalar :class:`gpuarray`.
    """
    out_t = dtype_to_ctype(dtype_z)

    # {{{ cast real scalars in context of complex scalar arithmetic

    if dtype_z.kind == "c" and dtype_x.kind != "c":
        dtype_z_real = _get_real_dtype(dtype_z)
        x_t = dtype_to_ctype(dtype_z_real)
    else:
        x_t = out_t

    if dtype_z.kind == "c" and dtype_y.kind != "c":
        dtype_z_real = _get_real_dtype(dtype_z)
        y_t = dtype_to_ctype(dtype_z_real)
    else:
        y_t = out_t

    # }}}

    x = "x[0]" if x_is_scalar else "x[i]"
    a = f"({x_t}) a"
    ax = f"{a}*(({x_t}) {x})"
    y = "y[0]" if y_is_scalar else "y[i]"
    b = f"({y_t}) b"
    by = f"({b})*(({y_t}) {y})"
    result = f"{ax} + {by}"
    return get_elwise_kernel(
        "%(tp_x)s a, %(tp_x)s *x, %(tp_y)s b, %(tp_y)s *y, %(tp_z)s *z"
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_y": dtype_to_ctype(dtype_y),
            "tp_z": dtype_to_ctype(dtype_z),
        },
        f"z[i] = {result}",
        "axpbyz",
    )


@context_dependent_memoize
def get_axpbz_kernel(dtype_x, dtype_z):
    return get_elwise_kernel(
        "%(tp_z)s a, %(tp_x)s *x,%(tp_z)s b, %(tp_z)s *z"
        % {"tp_x": dtype_to_ctype(dtype_x), "tp_z": dtype_to_ctype(dtype_z)},
        "z[i] = a * x[i] + b",
        "axpb",
    )


@context_dependent_memoize
def get_binary_op_kernel(dtype_x, dtype_y, dtype_z, operator,
                        x_is_scalar=False, y_is_scalar=False):
    """
    Returns a kernel corresponding to ``z = x (operator) y``.

    :arg x_is_scalar: A :class:`bool` which is *True* only if `x` is a scalar :class:`gpuarray`.
    :arg y_is_scalar: A :class:`bool` which is *True* only if `y` is a scalar :class:`gpuarray`.
    """

    out_t = dtype_to_ctype(dtype_z)

    # {{{ cast real scalars in context of complex scalar arithmetic

    if dtype_z.kind == "c" and dtype_x.kind != "c":
        dtype_z_real = _get_real_dtype(dtype_z)
        x_t = dtype_to_ctype(dtype_z_real)
    else:
        x_t = out_t

    if dtype_z.kind == "c" and dtype_y.kind != "c":
        dtype_z_real = _get_real_dtype(dtype_z)
        y_t = dtype_to_ctype(dtype_z_real)
    else:
        y_t = out_t

    # }}}

    x = "x[0]" if x_is_scalar else "x[i]"
    x = f"({x_t}) {x}"
    y = "y[0]" if y_is_scalar else "y[i]"
    y = f"({y_t}) {y}"
    result = f"{x} {operator} {y}"
    return get_elwise_kernel(
        "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z"
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_y": dtype_to_ctype(dtype_y),
            "tp_z": dtype_to_ctype(dtype_z),
        },
        f"z[i] = ({out_t}) {result}",
        "multiply",
    )


@context_dependent_memoize
def get_rdivide_elwise_kernel(dtype_x, dtype_z):
    return get_elwise_kernel(
        "%(tp_x)s *x, %(tp_z)s y, %(tp_z)s *z"
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_z": dtype_to_ctype(dtype_z),
        },
        "z[i] = y / x[i]",
        "divide_r",
    )


@context_dependent_memoize
def get_binary_func_kernel(func, dtype_x, dtype_y, dtype_z):
    return get_elwise_kernel(
        "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z"
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_y": dtype_to_ctype(dtype_y),
            "tp_z": dtype_to_ctype(dtype_z),
        },
        "z[i] = %s(x[i], y[i])" % func,
        func + "_kernel",
    )


@context_dependent_memoize
def get_binary_func_scalar_kernel(func, dtype_x, dtype_y, dtype_z):
    return get_elwise_kernel(
        "%(tp_x)s *x, %(tp_y)s y, %(tp_z)s *z"
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_y": dtype_to_ctype(dtype_y),
            "tp_z": dtype_to_ctype(dtype_z),
        },
        "z[i] = %s(x[i], y)" % func,
        func + "_kernel",
    )


def get_binary_minmax_kernel(func, dtype_x, dtype_y, dtype_z, use_scalar):
    if np.float64 not in [dtype_x, dtype_y]:
        func = func + "f"

    if any(dt.kind == "f" for dt in [dtype_x, dtype_y, dtype_z]):
        func = "f" + func

    if use_scalar:
        return get_binary_func_scalar_kernel(func, dtype_x, dtype_y, dtype_z)
    else:
        return get_binary_func_kernel(func, dtype_x, dtype_y, dtype_z)


@context_dependent_memoize
def get_fill_kernel(dtype):
    return get_elwise_kernel(
        "%(tp)s a, %(tp)s *z"
        % {
            "tp": dtype_to_ctype(dtype),
        },
        "z[i] = a",
        "fill",
    )


@context_dependent_memoize
def get_reverse_kernel(dtype):
    return get_elwise_kernel(
        "%(tp)s *y, %(tp)s *z"
        % {
            "tp": dtype_to_ctype(dtype),
        },
        "z[i] = y[n-1-i]",
        "reverse",
    )


@context_dependent_memoize
def get_real_kernel(dtype, real_dtype):
    return get_elwise_kernel(
        "%(tp)s *y, %(real_tp)s *z"
        % {
            "tp": dtype_to_ctype(dtype),
            "real_tp": dtype_to_ctype(real_dtype),
        },
        "z[i] = real(y[i])",
        "real",
    )


@context_dependent_memoize
def get_imag_kernel(dtype, real_dtype):
    return get_elwise_kernel(
        "%(tp)s *y, %(real_tp)s *z"
        % {
            "tp": dtype_to_ctype(dtype),
            "real_tp": dtype_to_ctype(real_dtype),
        },
        "z[i] = imag(y[i])",
        "imag",
    )


@context_dependent_memoize
def get_conj_kernel(dtype, conj_dtype):
    return get_elwise_kernel(
        "%(tp)s *y, %(conj_tp)s *z"
        % {
            "tp": dtype_to_ctype(dtype),
            "conj_tp": dtype_to_ctype(conj_dtype)
        },
        "z[i] = pycuda::conj(y[i])",
        "conj",
    )


@context_dependent_memoize
def get_arange_kernel(dtype):
    return get_elwise_kernel(
        "%(tp)s *z, %(tp)s start, %(tp)s step"
        % {
            "tp": dtype_to_ctype(dtype),
        },
        "z[i] = start + i*step",
        "arange",
    )


@context_dependent_memoize
def get_pow_array_kernel(dtype_x, dtype_y, dtype_z, is_base_array, is_exp_array):
    """
    Returns the kernel for the operation: ``z = x ** y``
    """
    if dtype_z == np.float32:
        func = "powf"
    else:
        # FIXME: Casting args to double-precision not
        # ideal for all cases (ex. int args)
        func = "pow"

    if not is_base_array and is_exp_array:
        x_ctype = "%(tp_x)s x"
        y_ctype = "%(tp_y)s *y"
        func = "%s(x,y[i])" % func

    elif is_base_array and is_exp_array:
        x_ctype = "%(tp_x)s *x"
        y_ctype = "%(tp_y)s *y"
        func = "%s(x[i],y[i])" % func

    elif is_base_array and not is_exp_array:
        x_ctype = "%(tp_x)s *x"
        y_ctype = "%(tp_y)s y"
        func = "%s(x[i],y)" % func

    else:
        raise AssertionError

    return get_elwise_kernel(
        (x_ctype + ", " + y_ctype + ", " + "%(tp_z)s *z")
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_y": dtype_to_ctype(dtype_y),
            "tp_z": dtype_to_ctype(dtype_z),
        },
        "z[i] = %s" % func,
        name="pow_method"
    )


@context_dependent_memoize
def get_fmod_kernel():
    return get_elwise_kernel(
        "float *arg, float *mod, float *z", "z[i] = fmod(arg[i], mod[i])", "fmod_kernel"
    )


@context_dependent_memoize
def get_modf_kernel():
    return get_elwise_kernel(
        "float *x, float *intpart ,float *fracpart",
        "fracpart[i] = modf(x[i], &intpart[i])",
        "modf_kernel",
    )


@context_dependent_memoize
def get_frexp_kernel():
    return get_elwise_kernel(
        "float *x, float *significand, float *exponent",
        """
                int expt = 0;
                significand[i] = frexp(x[i], &expt);
                exponent[i] = expt;
            """,
        "frexp_kernel",
    )


@context_dependent_memoize
def get_ldexp_kernel():
    return get_elwise_kernel(
        "float *sig, float *expt, float *z",
        "z[i] = ldexp(sig[i], int(expt[i]))",
        "ldexp_kernel",
    )


@context_dependent_memoize
def get_unary_func_kernel(func_name, in_dtype, out_dtype=None):
    if out_dtype is None:
        out_dtype = in_dtype

    return get_elwise_kernel(
        "%(tp_in)s *y, %(tp_out)s *z"
        % {
            "tp_in": dtype_to_ctype(in_dtype),
            "tp_out": dtype_to_ctype(out_dtype),
        },
        "z[i] = %s(y[i])" % func_name,
        "%s_kernel" % func_name,
    )


@context_dependent_memoize
def get_if_positive_kernel(crit_dtype, dtype):
    return get_elwise_kernel(
        [
            VectorArg(crit_dtype, "crit"),
            VectorArg(dtype, "then_"),
            VectorArg(dtype, "else_"),
            VectorArg(dtype, "result"),
        ],
        "result[i] = crit[i] > 0 ? then_[i] : else_[i]",
        "if_positive",
    )


@context_dependent_memoize
def get_where_kernel(crit_dtype, dtype):
    return get_elwise_kernel(
        [
            VectorArg(crit_dtype, "crit"),
            VectorArg(dtype, "then_"),
            VectorArg(dtype, "else_"),
            VectorArg(dtype, "result"),
        ],
        "result[i] = crit[i] != 0 ? then_[i] : else_[i]",
        "if_positive",
    )


@context_dependent_memoize
def get_scalar_op_kernel(dtype_x, dtype_a, dtype_y, operator):
    return get_elwise_kernel(
        "%(tp_x)s *x, %(tp_a)s a, %(tp_y)s *y"
        % {
            "tp_x": dtype_to_ctype(dtype_x),
            "tp_y": dtype_to_ctype(dtype_y),
            "tp_a": dtype_to_ctype(dtype_a),
        },
        "y[i] = x[i] %s a" % operator,
        "scalarop_kernel",
    )


@context_dependent_memoize
def get_logical_not_kernel(dtype_x, dtype_out):
    return get_elwise_kernel(
        [
            VectorArg(dtype_x, "x"),
            VectorArg(dtype_out, "out"),
        ],
        "out[i] = (x[i] == 0)",
        "logical_not",
    )



from __future__ import annotations

import numpy as np

from pytools import memoize_method

import pycuda.gpuarray as gpuarray
from pycuda.compiler import SourceModule


PKT_KERNEL_TEMPLATE = """
typedef %(index_type)s index_type;
typedef %(value_type)s value_type;
typedef %(packed_index_type)s packed_index_type;

#define ROWS_PER_PACKET %(rows_per_packet)d
#define THREADS_PER_PACKET %(threads_per_packet)d

template <typename IndexType, typename ValueType>
__device__ void memcpy_device(
  ValueType *dest, const ValueType *src,
  const IndexType num_values)
{
  for(unsigned int i = threadIdx.x; i < num_values; i += blockDim.x)
  {
    dest[i] = src[i];
  }
}

#define pkt_unpack_row_index(packed_index) ( packed_index >> 16  )
#define pkt_unpack_col_index(packed_index) (packed_index & 0xFFFF)

extern "C" {
__global__ void
spmv_pkt_kernel(const index_type *row_ptr,
                const index_type *pos_start,
                const index_type *pos_end,
                const packed_index_type *index_array,
                const value_type *data_array,
                const value_type *x,
                      value_type *y)
{
  __shared__ value_type s_x[ROWS_PER_PACKET]; // input x-values
  __shared__ value_type s_y[ROWS_PER_PACKET]; // output y-values

  const index_type thread_id =
    __umul24(THREADS_PER_PACKET, blockIdx.x) + threadIdx.x;

  // base index of the submatrix corresponding to this packet
  const index_type packet_base_row = row_ptr[blockIdx.x];
  const index_type packet_num_rows = row_ptr[blockIdx.x+1] - packet_base_row;

  // copy local x and y values from global memory into shared memory
  memcpy_device(s_x, x + packet_base_row, packet_num_rows);
  memcpy_device(s_y, y + packet_base_row, packet_num_rows);

  __syncthreads();

  // process packet

  const index_type packet_start = pos_start[thread_id];
  const index_type packet_end = pos_end[thread_id];

  for(index_type pos = packet_start; pos != packet_end; pos += THREADS_PER_PACKET)
  {
    // row and column indices are stored in the same 32-bit word

    const index_type packed_index = index_array[pos];

    const index_type row = pkt_unpack_row_index(packed_index);
    const index_type col = pkt_unpack_col_index(packed_index);
    const value_type val = data_array[pos];

    s_y[row] += val * s_x[col];
  }

  __syncthreads();

  // copy y-values from shared memory to global array

  memcpy_device(y + packet_base_row, s_y, packet_num_rows);
}
}
"""


class PacketedSpMV:
    def __init__(self, mat, is_symmetric, dtype):
        from pycuda.tools import DeviceData

        devdata = DeviceData()

        # all row indices in the data structure generation code are
        # "unpermuted" unless otherwise specified
        self.dtype = np.dtype(dtype)
        self.index_dtype = np.int32
        self.packed_index_dtype = np.uint32
        self.threads_per_packet = devdata.max_threads

        h, w = self.shape = mat.shape
        if h != w:
            raise ValueError("only square matrices are supported")

        self.rows_per_packet = (devdata.shared_memory - 100) // (
            2 * self.dtype.itemsize
        )

        self.block_count = (h + self.rows_per_packet - 1) // self.rows_per_packet

        # get metis partition -------------------------------------------------
        from scipy.sparse import csr_matrix

        csr_mat = csr_matrix(mat, dtype=self.dtype)

        from pymetis import part_graph

        if not is_symmetric:
            # make sure adjacency graph is undirected
            adj_mat = csr_mat + csr_mat.T
        else:
            adj_mat = csr_mat

        while True:
            _cut_count, dof_to_packet_nr = part_graph(
                int(self.block_count), xadj=adj_mat.indptr, adjncy=adj_mat.indices
            )

            # build packet_nr_to_dofs
            packet_nr_to_dofs = {}
            for i, packet_nr in enumerate(dof_to_packet_nr):
                try:
                    dof_packet = packet_nr_to_dofs[packet_nr]
                except KeyError:
                    packet_nr_to_dofs[packet_nr] = dof_packet = []

                dof_packet.append(i)

            packet_nr_to_dofs = [
                packet_nr_to_dofs.get(i) for i in range(len(packet_nr_to_dofs))
            ]

            too_big = False
            for packet_dofs in packet_nr_to_dofs:
                if len(packet_dofs) >= self.rows_per_packet:
                    too_big = True
                    break

            if too_big:
                old_block_count = self.block_count
                self.block_count = int(2 + 1.05 * self.block_count)
                print(

                        "Metis produced a big block at block count "
                        "%d--retrying with %d" % (old_block_count, self.block_count)

                )
                continue

            break

        assert len(packet_nr_to_dofs) == self.block_count

        # permutations, base rows ---------------------------------------------
        (
            new2old_fetch_indices,
            old2new_fetch_indices,
            packet_base_rows,
        ) = self.find_simple_index_stuff(packet_nr_to_dofs)

        # find local row cost and remaining_coo -------------------------------
        local_row_costs, remaining_coo = self.find_local_row_costs_and_remaining_coo(
            csr_mat, dof_to_packet_nr, old2new_fetch_indices
        )
        local_nnz = np.sum(local_row_costs)

        assert remaining_coo.nnz == csr_mat.nnz - local_nnz

        # find thread assignment for each block -------------------------------
        thread_count = len(packet_nr_to_dofs) * self.threads_per_packet
        thread_assignments, thread_costs = self.find_thread_assignment(
            packet_nr_to_dofs, local_row_costs, thread_count
        )

        max_thread_costs = np.max(thread_costs)

        # build data structure ------------------------------------------------
        from .pkt_build import build_pkt_data_structure

        build_pkt_data_structure(
            self,
            packet_nr_to_dofs,
            max_thread_costs,
            old2new_fetch_indices,
            csr_mat,
            thread_count,
            thread_assignments,
            local_row_costs,
        )

        self.packet_base_rows = gpuarray.to_gpu(packet_base_rows)
        self.new2old_fetch_indices = gpuarray.to_gpu(new2old_fetch_indices)
        self.old2new_fetch_indices = gpuarray.to_gpu(old2new_fetch_indices)

        from .coordinate import CoordinateSpMV

        self.remaining_coo_gpu = CoordinateSpMV(remaining_coo, dtype)

    def find_simple_index_stuff(self, packet_nr_to_dofs):
        new2old_fetch_indices = np.zeros(self.shape[0], dtype=self.index_dtype)
        old2new_fetch_indices = np.zeros(self.shape[0], dtype=self.index_dtype)

        packet_base_rows = np.zeros(self.block_count + 1, dtype=self.index_dtype)

        row_start = 0
        for packet_nr, packet in enumerate(packet_nr_to_dofs):
            packet_base_rows[packet_nr] = row_start
            row_end = row_start + len(packet)

            pkt_indices = np.array(packet, dtype=self.index_dtype)
            new2old_fetch_indices[row_start:row_end] = pkt_indices
            old2new_fetch_indices[pkt_indices] = np.arange(
                row_start, row_end, dtype=self.index_dtype
            )

            row_start += len(packet)

        packet_base_rows[self.block_count] = row_start

        return (new2old_fetch_indices, old2new_fetch_indices, packet_base_rows)

    def find_local_row_costs_and_remaining_coo(
        self, csr_mat, dof_to_packet_nr, old2new_fetch_indices
    ):
        h, _w = self.shape
        local_row_costs = [0] * h
        rem_coo_values = []
        rem_coo_i = []
        rem_coo_j = []

        iptr = csr_mat.indptr
        indices = csr_mat.indices
        data = csr_mat.data

        for i in range(h):
            for idx in range(iptr[i], iptr[i + 1]):
                j = indices[idx]

                if dof_to_packet_nr[i] == dof_to_packet_nr[j]:
                    local_row_costs[i] += 1
                else:
                    rem_coo_values.append(data[idx])
                    rem_coo_i.append(old2new_fetch_indices[i])
                    rem_coo_j.append(old2new_fetch_indices[j])

        from scipy.sparse import coo_matrix

        remaining_coo = coo_matrix(
            (rem_coo_values, (rem_coo_i, rem_coo_j)), self.shape, dtype=self.dtype
        )

        return local_row_costs, remaining_coo

    def find_thread_assignment(self, packet_nr_to_dofs, local_row_cost, thread_count):
        thread_assignments = [[] for i in range(thread_count)]
        thread_costs = np.zeros(thread_count)

        for packet_nr, packet_dofs in enumerate(packet_nr_to_dofs):
            row_costs_and_numbers = sorted(
                [(local_row_cost[i], i) for i in packet_dofs], reverse=True
            )

            base_thread_nr = packet_nr * self.threads_per_packet
            thread_offset = 0

            # zigzag assignment
            step = 1
            for row_cost, row_number in row_costs_and_numbers:
                ti = base_thread_nr + thread_offset
                thread_assignments[ti].append(row_number)
                thread_costs[ti] += row_cost

                if thread_offset + step >= self.threads_per_packet:
                    step = -1
                elif thread_offset + step < 0:
                    step = 1
                else:
                    thread_offset += step

        return thread_assignments, thread_costs

    def build_gpu_data_structure(
        self,
        packet_nr_to_dofs,
        max_thread_costs,
        old2new_fetch_indices,
        csr_mat,
        thread_count,
        thread_assignments,
        local_row_costs,
    ):
        # these arrays will likely be too long, but that's ok

        from .pkt_build import build_pkt_structure

        build_pkt_structure(
            self,
            packet_nr_to_dofs,
            thread_assignments,
            # thread_starts,
            # thread_ends,
            # index_array,
            # data_array,
        )

        # copy data to the gpu ------------------------------------------------

    # execution ---------------------------------------------------------------
    @memoize_method
    def get_kernel(self):
        from pycuda.tools import dtype_to_ctype

        mod = SourceModule(
            PKT_KERNEL_TEMPLATE
            % {
                "value_type": dtype_to_ctype(self.dtype),
                "index_type": dtype_to_ctype(self.index_dtype),
                "packed_index_type": dtype_to_ctype(self.packed_index_dtype),
                "threads_per_packet": self.threads_per_packet,
                "rows_per_packet": self.rows_per_packet,
            },
            no_extern_c=True,
        )
        func = mod.get_function("spmv_pkt_kernel")
        func.prepare("PPPPPPP")
        return func

    def permute(self, x):
        return gpuarray.take(x, self.new2old_fetch_indices)

    def unpermute(self, x):
        return gpuarray.take(x, self.old2new_fetch_indices)

    def __call__(self, x, y=None):
        if y is None:
            y = gpuarray.zeros(self.shape[0], dtype=self.dtype, allocator=x.allocator)

        self.get_kernel().prepared_call(
            (self.block_count, 1),
            (self.threads_per_packet, 1, 1),
            self.packet_base_rows.gpudata,
            self.thread_starts.gpudata,
            self.thread_ends.gpudata,
            self.index_array.gpudata,
            self.data_array.gpudata,
            x.gpudata,
            y.gpudata,
        )

        self.remaining_coo_gpu(x, y)

        return y



from __future__ import annotations


class OperatorBase:
    @property
    def dtype(self):
        raise NotImplementedError

    @property
    def shape(self):
        raise NotImplementedError


class IdentityOperator(OperatorBase):
    def __init__(self, dtype, n):
        self.my_dtype = dtype
        self.n = n

    @property
    def dtype(self):
        return self.my_dtype

    @property
    def shape(self):
        return self.n, self.n

    def __call__(self, operand):
        return operand


class DiagonalPreconditioner(OperatorBase):
    def __init__(self, diagonal):
        self.diagonal = diagonal

    @property
    def dtype(self):
        return self.diagonal.dtype

    @property
    def shape(self):
        n = self.diagonal.shape[0]
        return n, n

    def __call__(self, operand):
        return self.diagonal * operand



from __future__ import annotations

import sys
from optparse import OptionParser
from os.path import exists

import pycuda.driver


pycuda.driver.set_debugging()

parser = OptionParser(usage="usage: %prog [options] SCRIPT-TO-RUN [SCRIPT-ARGUMENTS]")

parser.disable_interspersed_args()
options, args = parser.parse_args()

if len(args) < 1:
    parser.print_help()
    sys.exit(2)

mainpyfile = args[0]

if not exists(mainpyfile):
    print("Error:", mainpyfile, "does not exist")
    sys.exit(1)

sys.argv = args

with open(mainpyfile) as mainpy:
    exec(compile(mainpy.read(), mainpyfile, "exec"))



"""Computation of reductions on vectors."""
from __future__ import annotations


__copyright__ = "Copyright (C) 2009 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Based on code/ideas by Mark Harris <mharris@nvidia.com>.

Original License:

Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.

NOTICE TO USER:

This source code is subject to NVIDIA ownership rights under U.S. and
international Copyright laws.

NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
OR PERFORMANCE OF THIS SOURCE CODE.

U.S. Government End Users.  This source code is a "commercial item" as
that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
"commercial computer software" and "commercial computer software
documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
and is provided to the U.S. Government only as a commercial end item.
Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
source code with only those rights set forth herein.
"""

import numpy as np

from pycuda.tools import context_dependent_memoize, dtype_to_ctype


def get_reduction_module(
    out_type,
    block_size,
    neutral,
    reduce_expr,
    map_expr,
    arguments,
    name="reduce_kernel",
    keep=False,
    options=None,
    preamble="",
):

    from pycuda.compiler import SourceModule

    src = """
        #include <pycuda-complex.hpp>

        #define BLOCK_SIZE %(block_size)d
        #define READ_AND_MAP(i) (%(map_expr)s)
        #define REDUCE(a, b) (%(reduce_expr)s)

        %(preamble)s

        typedef %(out_type)s out_type;

        extern "C"
        __global__
        void %(name)s(out_type *out, %(arguments)s,
          unsigned int seq_count, size_t n)
        {
          // Needs to be variable-size to prevent the braindead CUDA compiler from
          // running constructors on this array. Grrrr.
          extern __shared__ out_type sdata[];

          unsigned int tid = threadIdx.x;

          size_t i = blockIdx.x*BLOCK_SIZE*seq_count + tid;

          out_type acc = %(neutral)s;
          for (unsigned s = 0; s < seq_count; ++s)
          {
            if (i >= n)
              break;
            acc = REDUCE(acc, READ_AND_MAP(i));

            i += BLOCK_SIZE;
          }

          sdata[tid] = acc;

          __syncthreads();

          #if (BLOCK_SIZE >= 512)
            if (tid < 256) { sdata[tid] = REDUCE(sdata[tid], sdata[tid + 256]); }
            __syncthreads();
          #endif

          #if (BLOCK_SIZE >= 256)
            if (tid < 128) { sdata[tid] = REDUCE(sdata[tid], sdata[tid + 128]); }
            __syncthreads();
          #endif

          #if (BLOCK_SIZE >= 128)
            if (tid < 64) { sdata[tid] = REDUCE(sdata[tid], sdata[tid + 64]); }
            __syncthreads();
          #endif

          if (tid < 32)
          {
            // 'volatile' required according to Fermi compatibility guide 1.2.2
            volatile out_type *smem = sdata;
            if (BLOCK_SIZE >= 64) smem[tid] = REDUCE(smem[tid], smem[tid + 32]);
            if (BLOCK_SIZE >= 32) smem[tid] = REDUCE(smem[tid], smem[tid + 16]);
            if (BLOCK_SIZE >= 16) smem[tid] = REDUCE(smem[tid], smem[tid + 8]);
            if (BLOCK_SIZE >= 8)  smem[tid] = REDUCE(smem[tid], smem[tid + 4]);
            if (BLOCK_SIZE >= 4)  smem[tid] = REDUCE(smem[tid], smem[tid + 2]);
            if (BLOCK_SIZE >= 2)  smem[tid] = REDUCE(smem[tid], smem[tid + 1]);
          }

          if (tid == 0) out[blockIdx.x] = sdata[0];
        }
        """ % {
        "out_type": out_type,
        "arguments": arguments,
        "block_size": block_size,
        "neutral": neutral,
        "reduce_expr": reduce_expr,
        "map_expr": map_expr,
        "name": name,
        "preamble": preamble,
    }
    return SourceModule(src, options=options, keep=keep, no_extern_c=True)


def get_reduction_kernel_and_types(
    stage,
    out_type,
    block_size,
    neutral,
    reduce_expr,
    map_expr=None,
    arguments=None,
    name="reduce_kernel",
    keep=False,
    options=None,
    preamble="",
):

    if stage == 1:
        if map_expr is None:
            map_expr = "in[i]"

    elif stage == 2:
        if map_expr is None:
            map_expr = "pycuda_reduction_inp[i]"

        in_arg = "const %s *pycuda_reduction_inp" % out_type
        arguments = in_arg + ", " + arguments if arguments else in_arg

    else:
        raise AssertionError()

    mod = get_reduction_module(
        out_type,
        block_size,
        neutral,
        reduce_expr,
        map_expr,
        arguments,
        name,
        keep,
        options,
        preamble,
    )

    from pycuda.tools import get_arg_type

    func = mod.get_function(name)
    arg_types = [get_arg_type(arg) for arg in arguments.split(",")]
    func.prepare("P%sIN" % "".join(arg_types))

    return func, arg_types


class ReductionKernel:
    def __init__(
        self,
        dtype_out,
        neutral,
        reduce_expr,
        map_expr=None,
        arguments=None,
        name="reduce_kernel",
        keep=False,
        options=None,
        preamble="",
    ):

        self.dtype_out = np.dtype(dtype_out)

        self.block_size = 512

        s1_func, self.stage1_arg_types = get_reduction_kernel_and_types(
            1,
            dtype_to_ctype(dtype_out),
            self.block_size,
            neutral,
            reduce_expr,
            map_expr,
            arguments,
            name=name + "_stage1",
            keep=keep,
            options=options,
            preamble=preamble,
        )
        self.stage1_func = s1_func.prepared_async_call

        # stage 2 has only one input and no map expression
        s2_func, self.stage2_arg_types = get_reduction_kernel_and_types(
            2,
            dtype_to_ctype(dtype_out),
            self.block_size,
            neutral,
            reduce_expr,
            arguments=arguments,
            name=name + "_stage2",
            keep=keep,
            options=options,
            preamble=preamble,
        )
        self.stage2_func = s2_func.prepared_async_call

        assert [i for i, arg_tp in enumerate(self.stage1_arg_types) if arg_tp == "P"], (
            "ReductionKernel can only be used with functions that have at least one "
            "vector argument"
        )

    def __call__(self, *args, **kwargs):
        MAX_BLOCK_COUNT = 1024  # noqa: N806
        SMALL_SEQ_COUNT = 4  # noqa: N806

        s1_func = self.stage1_func
        s2_func = self.stage2_func

        kernel_wrapper = kwargs.get("kernel_wrapper")
        if kernel_wrapper is not None:
            s1_func = kernel_wrapper(s1_func)
            s2_func = kernel_wrapper(s2_func)

        stream = kwargs.get("stream")
        out = kwargs.pop("out", None)

        from .gpuarray import empty

        f = s1_func
        arg_types = self.stage1_arg_types

        stage1_args = args

        while True:
            invocation_args = []
            vectors = []

            for arg, arg_tp in zip(args, arg_types, strict=False):
                if arg_tp == "P":
                    if not arg.flags.forc:
                        raise RuntimeError(
                            "ReductionKernel cannot " "deal with non-contiguous arrays"
                        )

                    vectors.append(arg)
                    invocation_args.append(arg.gpudata)
                else:
                    invocation_args.append(arg)

            repr_vec = vectors[0]
            sz = repr_vec.size

            allocator = kwargs.get("allocator")
            if allocator is None:
                allocator = repr_vec.allocator

            if sz <= self.block_size * SMALL_SEQ_COUNT * MAX_BLOCK_COUNT:
                total_block_size = SMALL_SEQ_COUNT * self.block_size
                block_count = (sz + total_block_size - 1) // total_block_size
                seq_count = SMALL_SEQ_COUNT
            else:
                block_count = MAX_BLOCK_COUNT
                macroblock_size = block_count * self.block_size
                seq_count = (sz + macroblock_size - 1) // macroblock_size

            if block_count == 1 and out is not None:
                if out.dtype != self.dtype_out:
                    raise ValueError("out must have the same dtype as dtype_out")
                if out.size == 0:
                    raise ValueError("out array is empty")
                result = out
            elif block_count == 1:
                result = empty((), self.dtype_out, allocator=allocator)
            else:
                result = empty((block_count,), self.dtype_out, allocator=allocator)

            kwargs = {"shared_size": self.block_size * self.dtype_out.itemsize}

            # print block_count, seq_count, self.block_size, sz
            f(
                (block_count, 1),
                (self.block_size, 1, 1),
                stream,
                *([result.gpudata, *invocation_args, seq_count, sz]),
                **kwargs
            )

            if block_count == 1:
                return result
            else:
                f = s2_func
                arg_types = self.stage2_arg_types
                args = (result, *stage1_args)


@context_dependent_memoize
def get_sum_kernel(dtype_out, dtype_in):
    if dtype_out is None:
        dtype_out = dtype_in

    return ReductionKernel(
        dtype_out,
        "0",
        "a+b",
        arguments="const {tp} *in".format(tp=dtype_to_ctype(dtype_in)),
    )


@context_dependent_memoize
def get_any_kernel(dtype_out, dtype_in):
    if dtype_out is None:
        dtype_out = dtype_in

    return ReductionKernel(
        dtype_out,
        "0",
        "(a != 0) || (b != 0)",
        arguments="const {tp} *in".format(tp=dtype_to_ctype(dtype_in)),
    )


@context_dependent_memoize
def get_all_kernel(dtype_out, dtype_in):
    if dtype_out is None:
        dtype_out = dtype_in

    return ReductionKernel(
        dtype_out,
        "1",
        "(a != 0) && (b != 0)",
        arguments="const {tp} *in".format(tp=dtype_to_ctype(dtype_in)),
    )


@context_dependent_memoize
def get_subset_sum_kernel(dtype_out, dtype_subset, dtype_in):
    if dtype_out is None:
        dtype_out = dtype_in

    return ReductionKernel(
        dtype_out,
        "0",
        "a+b",
        map_expr="in[lookup_tbl[i]]",
        arguments="const %(tp_lut)s *lookup_tbl, const %(tp)s *in"
        % {
            "tp": dtype_to_ctype(dtype_in),
            "tp_lut": dtype_to_ctype(dtype_subset),
        },
    )


@context_dependent_memoize
def get_dot_kernel(dtype_out, dtype_a, dtype_b):
    return ReductionKernel(
        dtype_out,
        neutral="0",
        reduce_expr="a+b",
        map_expr="a[i]*b[i]",
        arguments="const %(tp_a)s *a, const %(tp_b)s *b"
        % {
            "tp_a": dtype_to_ctype(dtype_a),
            "tp_b": dtype_to_ctype(dtype_b),
        },
        keep=True,
    )


@context_dependent_memoize
def get_subset_dot_kernel(dtype_out, dtype_subset, dtype_a=None, dtype_b=None):
    if dtype_out is None:
        dtype_out = dtype_a

    if dtype_b is None:
        dtype_b = dtype_out if dtype_a is None else dtype_a

    if dtype_a is None:
        dtype_a = dtype_out

    # important: lookup_tbl must be first--it controls the length
    return ReductionKernel(
        dtype_out,
        neutral="0",
        reduce_expr="a+b",
        map_expr="a[lookup_tbl[i]]*b[lookup_tbl[i]]",
        arguments="const %(tp_lut)s *lookup_tbl, "
        "const %(tp_a)s *a, const %(tp_b)s *b"
        % {
            "tp_a": dtype_to_ctype(dtype_a),
            "tp_b": dtype_to_ctype(dtype_b),
            "tp_lut": dtype_to_ctype(dtype_subset),
        },
    )


def get_minmax_neutral(what, dtype):
    dtype = np.dtype(dtype)
    if issubclass(dtype.type, np.inexact):
        if what == "min":
            return "MY_INFINITY"
        elif what == "max":
            return "-MY_INFINITY"
        else:
            raise ValueError("what is not min or max.")
    else:
        if what == "min":
            return str(np.iinfo(dtype).max)
        elif what == "max":
            return str(np.iinfo(dtype).min)
        else:
            raise ValueError("what is not min or max.")


@context_dependent_memoize
def get_minmax_kernel(what, dtype):
    if dtype == np.float64:
        reduce_expr = "f%s(a,b)" % what
    elif dtype == np.float32:
        reduce_expr = "f%sf(a,b)" % what
    elif dtype.kind in "iu":
        reduce_expr = "%s(a,b)" % what
    else:
        raise TypeError("unsupported dtype specified")

    return ReductionKernel(
        dtype,
        neutral=get_minmax_neutral(what, dtype),
        reduce_expr=f"{reduce_expr}",
        arguments="const %(tp)s *in"
        % {
            "tp": dtype_to_ctype(dtype),
        },
        preamble="#define MY_INFINITY (1./0)",
    )


@context_dependent_memoize
def get_subset_minmax_kernel(what, dtype, dtype_subset):
    if dtype == np.float64:
        reduce_expr = "f%s(a,b)" % what
    elif dtype == np.float32:
        reduce_expr = "f%sf(a,b)" % what
    elif dtype.kind in "iu":
        reduce_expr = "%s(a,b)" % what
    else:
        raise TypeError("unsupported dtype specified")

    return ReductionKernel(
        dtype,
        neutral=get_minmax_neutral(what, dtype),
        reduce_expr=f"{reduce_expr}",
        map_expr="in[lookup_tbl[i]]",
        arguments="const %(tp_lut)s *lookup_tbl, "
        "const %(tp)s *in"
        % {
            "tp": dtype_to_ctype(dtype),
            "tp_lut": dtype_to_ctype(dtype_subset),
        },
        preamble="#define MY_INFINITY (1./0)",
    )



# don't import pycuda.driver here--you'll create an import loop
from __future__ import annotations

import os
import sys
from os import unlink
from tempfile import mkstemp

from pytools import memoize
from pytools.prefork import call_capture_output


@memoize
def get_nvcc_version(nvcc):
    cmdline = [nvcc, "--version"]
    result, stdout, _stderr = call_capture_output(cmdline)

    if result != 0 or not stdout:
        from warnings import warn

        warn("NVCC version could not be determined.", stacklevel=2)
        stdout = b"nvcc unknown version"

    return stdout.decode("utf-8", "replace")


def _new_md5():
    try:
        import hashlib

        return hashlib.md5()
    except ImportError:
        # for Python << 2.5
        import md5

        return md5.new()


def preprocess_source(source, options, nvcc):
    handle, source_path = mkstemp(suffix=".cu")

    with open(source_path, "w") as outf:
        outf.write(source)

    os.close(handle)

    cmdline = [nvcc, "--preprocess", *options, source_path]
    if "win32" in sys.platform:
        cmdline.extend(["--compiler-options", "-EP"])
    else:
        cmdline.extend(["--compiler-options", "-P"])

    result, stdout, stderr = call_capture_output(cmdline, error_on_nonzero=False)

    if result != 0:
        from pycuda.driver import CompileError

        raise CompileError(
            "nvcc preprocessing of %s failed" % source_path, cmdline, stderr=stderr
        )

    # sanity check
    if len(stdout) < 0.5 * len(source):
        from pycuda.driver import CompileError

        raise CompileError(
            "nvcc preprocessing of %s failed with ridiculously "
            "small code output - likely unsupported compiler." % source_path,
            cmdline,
            stderr=stderr.decode("utf-8", "replace"),
        )

    unlink(source_path)

    preprocessed_str = stdout.decode("utf-8", "replace")

    # remove the temporary filename from the preprocessed source code to get reproducible hashes
    return preprocessed_str.replace(os.path.basename(source_path), "")


def compile_plain(source, options, keep, nvcc, cache_dir, target="cubin"):
    from os.path import join

    assert target in ["cubin", "ptx", "fatbin"]

    if cache_dir:
        checksum = _new_md5()

        if "#include" in source:
            checksum.update(preprocess_source(source, options, nvcc).encode("utf-8"))
        else:
            checksum.update(source.encode("utf-8"))

        for option in options:
            checksum.update(option.encode("utf-8"))
        checksum.update(get_nvcc_version(nvcc).encode("utf-8"))
        from pycuda.characterize import platform_bits

        checksum.update(str(platform_bits()).encode("utf-8"))

        cache_file = checksum.hexdigest()
        cache_path = join(cache_dir, cache_file + "." + target)

        try:
            with open(cache_path, "rb") as cache_file:
                return cache_file.read()

        except Exception:
            pass

    from tempfile import mkdtemp

    file_dir = mkdtemp()
    file_root = "kernel"

    cu_file_name = file_root + ".cu"
    cu_file_path = join(file_dir, cu_file_name)

    with open(cu_file_path, "w") as outf:
        outf.write(str(source))

    if keep:
        options = options[:]
        options.append("--keep")

        print("*** compiler output in %s" % file_dir)

    cmdline = [nvcc, "--" + target, *options, cu_file_name]
    result, stdout, stderr = call_capture_output(
        cmdline, cwd=file_dir, error_on_nonzero=False
    )

    try:
        result_f = open(join(file_dir, file_root + "." + target), "rb")  # noqa: SIM115
    except OSError:
        no_output = True
    else:
        no_output = False

    if result != 0 or (no_output and (stdout or stderr)):
        if result == 0:
            from warnings import warn

            warn(
                "PyCUDA: nvcc exited with status 0, but appears to have "
                "encountered an error", stacklevel=2
            )
        from pycuda.driver import CompileError

        raise CompileError(
            "nvcc compilation of %s failed" % cu_file_path,
            cmdline,
            stdout=stdout.decode("utf-8", "replace"),
            stderr=stderr.decode("utf-8", "replace"),
        )

    if stdout or stderr:
        lcase_err_text = (stdout + stderr).decode("utf-8", "replace").lower()
        from warnings import warn

        if "demoted" in lcase_err_text or "demoting" in lcase_err_text:
            warn(
                "nvcc said it demoted types in source code it "
                "compiled--this is likely not what you want.",
                stacklevel=4,
            )
        warn(
            "The CUDA compiler succeeded, but said the following:\n"
            + (stdout + stderr).decode("utf-8", "replace"),
            stacklevel=4,
        )

    result_data = result_f.read()
    result_f.close()

    if cache_dir:
        with open(cache_path, "wb") as outf:
            outf.write(result_data)

    if not keep:
        from os import listdir, rmdir, unlink

        for name in listdir(file_dir):
            unlink(join(file_dir, name))
        rmdir(file_dir)

    return result_data


def _get_per_user_string():
    try:
        from os import getuid
    except ImportError:
        checksum = _new_md5()
        from os import environ

        checksum.update(environ["USERNAME"].encode("utf-8"))
        return checksum.hexdigest()
    else:
        return "uid%d" % getuid()


def _find_pycuda_include_path():
    import importlib.util
    import os

    return os.path.abspath(
        os.path.join(importlib.util.find_spec("pycuda").origin,
                     os.path.pardir, "cuda"))


DEFAULT_NVCC_FLAGS = [
    flag.strip()
    for flag in os.environ.get("PYCUDA_DEFAULT_NVCC_FLAGS", "").split()
    if flag.strip()
]


def compile(
    source,
    nvcc="nvcc",
    options=None,
    keep=False,
    no_extern_c=False,
    arch=None,
    code=None,
    cache_dir=None,
    include_dirs=None,
    target="cubin",
):

    if include_dirs is None:
        include_dirs = []
    assert target in ["cubin", "ptx", "fatbin"]

    if not no_extern_c:
        source = 'extern "C" {\n%s\n}\n' % source

    if options is None:
        options = DEFAULT_NVCC_FLAGS

    options = options[:]
    if arch is None:
        from pycuda.driver import Error

        try:
            from pycuda.driver import Context

            arch = "sm_%d%d" % Context.get_device().compute_capability()
        except Error:
            pass

    from pycuda.driver import CUDA_DEBUGGING

    if CUDA_DEBUGGING:
        cache_dir = False
        keep = True
        options.extend(["-g", "-G"])

    if "PYCUDA_CACHE_DIR" in os.environ and cache_dir is None:
        cache_dir = os.environ["PYCUDA_CACHE_DIR"]

    if "PYCUDA_DISABLE_CACHE" in os.environ:
        cache_dir = False

    if cache_dir is None:
        import platformdirs

        cache_dir = os.path.join(
            platformdirs.user_cache_dir("pycuda", "pycuda"), "compiler-cache-v1"
        )

        from os import makedirs
        makedirs(cache_dir, exist_ok=True)

    if arch is not None:
        options.extend(["-arch", arch])

    if code is not None:
        options.extend(["-code", code])

    if (
            ("darwin" in sys.platform and sys.maxsize == 9223372036854775807)
            or
            ("win32" in sys.platform and sys.maxsize == 9223372036854775807)):
        options.append("-m64")
    elif "win32" in sys.platform and sys.maxsize == 2147483647:
        options.append("-m32")

    include_dirs = [*include_dirs, _find_pycuda_include_path()]

    for i in include_dirs:
        options.append("-I" + i)

    return compile_plain(source, options, keep, nvcc, cache_dir, target)


class CudaModule:
    def _check_arch(self, arch):
        if arch is None:
            return
        try:
            from pycuda.driver import Context

            capability = Context.get_device().compute_capability()
            if tuple(map(int, tuple(arch.split("_")[1]))) > capability:
                from warnings import warn

                warn(
                    "trying to compile for a compute capability "
                    "higher than selected GPU", stacklevel=2
                )
        except Exception:
            pass

    def _bind_module(self):
        self.get_global = self.module.get_global
        self.get_texref = self.module.get_texref
        if hasattr(self.module, "get_surfref"):
            self.get_surfref = self.module.get_surfref

    def get_function(self, name):
        return self.module.get_function(name)


class SourceModule(CudaModule):
    """
    Creates a Module from a single .cu source object linked against the
    static CUDA runtime.
    """

    def __init__(
        self,
        source,
        nvcc="nvcc",
        options=None,
        keep=False,
        no_extern_c=False,
        arch=None,
        code=None,
        cache_dir=None,
        include_dirs=None,
    ):
        if include_dirs is None:
            include_dirs = []
        self._check_arch(arch)

        cubin = compile(
            source,
            nvcc,
            options,
            keep,
            no_extern_c,
            arch,
            code,
            cache_dir,
            include_dirs,
        )

        from pycuda.driver import module_from_buffer

        self.module = module_from_buffer(cubin)

        self._bind_module()


def _search_on_path(filenames):
    """Find file on system path."""
    # http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52224

    from os import environ, pathsep
    from os.path import abspath, exists, join

    search_path = environ["PATH"]

    paths = search_path.split(pathsep)
    for path in paths:
        for filename in filenames:
            if exists(join(path, filename)):
                return abspath(join(path, filename))


@memoize
def _find_nvcc_on_path():
    return _search_on_path(["nvcc", "nvcc.exe"])


class DynamicModule(CudaModule):
    """
    Creates a Module from multiple .cu source, library file and/or data
    objects linked against the static or dynamic CUDA runtime.
    """

    def __init__(
        self,
        nvcc="nvcc",
        link_options=None,
        keep=False,
        no_extern_c=False,
        arch=None,
        code=None,
        cache_dir=None,
        include_dirs=None,
        message_handler=None,
        log_verbose=False,
        cuda_libdir=None,
    ):
        from pycuda.driver import Context

        if include_dirs is None:
            include_dirs = []
        compute_capability = Context.get_device().compute_capability()
        if compute_capability < (3, 5):
            raise Exception(
                "Minimum compute capability for dynamic parallelism is 3.5 (found: %u.%u)!"
                % (compute_capability[0], compute_capability[1])
            )
        else:
            from pycuda.driver import Linker

            self.linker = Linker(message_handler, link_options, log_verbose)
        self._check_arch(arch)
        self.nvcc = nvcc
        self.keep = keep
        self.no_extern_c = no_extern_c
        self.arch = arch
        self.code = code
        self.cache_dir = cache_dir
        self.include_dirs = include_dirs
        self.cuda_libdir = cuda_libdir
        self.libdir, self.libptn = None, None
        self.module = None

    def _locate_cuda_libdir(self):
        """
        Locate the "standard" CUDA SDK library directory in the local
        file system. Supports 64-Bit Windows, Linux and Mac OS X.
        In case the caller supplied cuda_libdir in the constructor
        other than None that value is returned unchecked, else a
        best-effort attempt is made.
        Precedence:
            Windows: cuda_libdir > %CUDA_PATH%
            Linux:   cuda_libdir > $CUDA_ROOT > $LD_LIBRARY_PATH > '/usr/lib/x86_64-linux-gnu'
        Returns a pair (libdir, libptn) where libdir is None in case
            of failure or a string containing the absolute path of the
            directory, and libptn is the %-format pattern to construct
            library file names from library names on the local system.
        Raises a RuntimeError in case of failure.
        Links:
        - Post-installation Actions
          http://docs.nvidia.com/cuda/cuda-installation-guide-linux/index.html#post-installation-actions
        TODO:
        - Is $CUDA_ROOT/lib64 the correct path to assume for 64-Bit CUDA libraries on Linux?
        - Mac OS X (Darwin) is currently treated like Linux, is that correct?
        - Check CMake's FindCUDA module, it might contain some helpful clues in its sources
          https://cmake.org/cmake/help/v3.0/module/FindCUDA.html
          https://github.com/Kitware/CMake/blob/master/Modules/FindCUDA.cmake
        - Verify all Linux code paths somehow
        """
        from os.path import isfile, join
        from platform import system as platform_system

        system = platform_system()
        libdir, libptn = None, None
        if system == "Windows":
            if self.cuda_libdir is not None:
                libdir = self.cuda_libdir
            elif "CUDA_PATH" in os.environ and isfile(
                join(os.environ["CUDA_PATH"], "lib\\x64\\cudadevrt.lib")
            ):
                libdir = join(os.environ["CUDA_PATH"], "lib\\x64")
            libptn = "%s.lib"
        elif system in ["Linux", "Darwin"]:
            if self.cuda_libdir is not None:
                libdir = self.cuda_libdir
            elif "CUDA_ROOT" in os.environ and isfile(
                join(os.environ["CUDA_ROOT"], "lib64/libcudadevrt.a")
            ):
                libdir = join(os.environ["CUDA_ROOT"], "lib64")
            elif "LD_LIBRARY_PATH" in os.environ:
                for ld_path in os.environ["LD_LIBRARY_PATH"].split(":"):
                    if isfile(join(ld_path, "libcudadevrt.a")):
                        libdir = ld_path
                        break

            if libdir is None and isfile("/usr/lib/x86_64-linux-gnu/libcudadevrt.a"):
                libdir = "/usr/lib/x86_64-linux-gnu"

            if libdir is None:
                nvcc_path = _find_nvcc_on_path()
                if nvcc_path is not None:
                    libdir = join(os.path.dirname(nvcc_path), "..", "lib64")

            libptn = "lib%s.a"
        if libdir is None:
            raise RuntimeError(
                "Unable to locate the CUDA SDK installation "
                "directory, set CUDA library path manually"
            )
        return libdir, libptn

    def add_source(self, source, nvcc_options=None, name="kernel.ptx"):
        ptx = compile(
            source,
            nvcc=self.nvcc,
            options=nvcc_options,
            keep=self.keep,
            no_extern_c=self.no_extern_c,
            arch=self.arch,
            code=self.code,
            cache_dir=self.cache_dir,
            include_dirs=self.include_dirs,
            target="ptx",
        )
        from pycuda.driver import jit_input_type

        self.linker.add_data(ptx, jit_input_type.PTX, name)
        return self

    def add_data(self, data, input_type, name="unknown"):
        self.linker.add_data(data, input_type, name)
        return self

    def add_file(self, filename, input_type):
        self.linker.add_file(filename, input_type)
        return self

    def add_stdlib(self, libname):
        if self.libdir is None:
            self.libdir, self.libptn = self._locate_cuda_libdir()
        from os.path import isfile, join

        libpath = join(self.libdir, self.libptn % libname)
        if not isfile(libpath):
            raise OSError('CUDA SDK library file "%s" not found' % libpath)
        from pycuda.driver import jit_input_type

        self.linker.add_file(libpath, jit_input_type.LIBRARY)
        return self

    def link(self):
        self.module = self.linker.link_module()
        self.linker = None
        self._bind_module()
        return self


class DynamicSourceModule(DynamicModule):
    """
    Creates a Module from a single .cu source object linked against the
    dynamic CUDA runtime.
    - compiler generates PTX relocatable device code (rdc) from source that
      can be linked with other relocatable device code
    - source is linked against the CUDA device runtime library cudadevrt
    - library cudadevrt is statically linked into the generated Module
    """

    def __init__(
        self,
        source,
        nvcc="nvcc",
        options=None,
        keep=False,
        no_extern_c=False,
        arch=None,
        code=None,
        cache_dir=None,
        include_dirs=None,
        cuda_libdir=None,
    ):
        if include_dirs is None:
            include_dirs = []
        super().__init__(
            nvcc=nvcc,
            link_options=None,
            keep=keep,
            no_extern_c=no_extern_c,
            arch=arch,
            code=code,
            cache_dir=cache_dir,
            include_dirs=include_dirs,
            cuda_libdir=cuda_libdir,
        )
        if options is None:
            options = DEFAULT_NVCC_FLAGS
        options = options[:]
        if "-rdc=true" not in options:
            options.append("-rdc=true")
        if "-lcudadevrt" not in options:
            options.append("-lcudadevrt")
        self.add_source(source, nvcc_options=options)
        self.add_stdlib("cudadevrt")
        self.link()



from __future__ import annotations

import numpy as np

from pytools import memoize_method

import pycuda.driver as drv
import pycuda.gpuarray as gpuarray
from pycuda.sparse.inner import AsyncInnerProduct


class ConvergenceError(RuntimeError):
    pass


class CGStateContainer:
    def __init__(self, operator, precon=None, pagelocked_allocator=None):
        if precon is None:
            from pycuda.sparse.operator import IdentityOperator

            precon = IdentityOperator(operator.dtype, operator.shape[0])

        self.operator = operator
        self.precon = precon

        self.pagelocked_allocator = pagelocked_allocator

    @memoize_method
    def make_lc2_kernel(self, dtype, a_is_gpu, b_is_gpu):
        from pycuda.elementwise import get_linear_combination_kernel

        return get_linear_combination_kernel(
            ((a_is_gpu, dtype, dtype), (b_is_gpu, dtype, dtype)), dtype
        )

    def lc2(self, a, x, b, y, out=None):
        if out is None:
            out = gpuarray.empty(x.shape, dtype=x.dtype, allocator=x.allocator)

        assert x.dtype == y.dtype == out.dtype
        a_is_gpu = isinstance(a, gpuarray.GPUArray)
        b_is_gpu = isinstance(b, gpuarray.GPUArray)
        assert x.shape == y.shape == out.shape

        kernel, texrefs = self.make_lc2_kernel(x.dtype, a_is_gpu, b_is_gpu)

        texrefs = texrefs[:]

        args = []

        if a_is_gpu:
            assert a.dtype == x.dtype
            assert a.shape == ()
            a.bind_to_texref_ext(texrefs.pop(0), allow_double_hack=True)
        else:
            args.append(a)
        args.append(x.gpudata)

        if b_is_gpu:
            assert b.dtype == y.dtype
            assert b.shape == ()
            b.bind_to_texref_ext(texrefs.pop(0), allow_double_hack=True)
        else:
            args.append(b)
        args.append(y.gpudata)
        args.append(out.gpudata)
        args.append(x.mem_size)

        kernel.prepared_call(x._grid, x._block, *args)

        return out

    @memoize_method
    def guarded_div_kernel(self, dtype_x, dtype_y, dtype_z):
        from pycuda.elementwise import get_elwise_kernel
        from pycuda.tools import dtype_to_ctype

        return get_elwise_kernel(
            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z"
            % {
                "tp_x": dtype_to_ctype(dtype_x),
                "tp_y": dtype_to_ctype(dtype_y),
                "tp_z": dtype_to_ctype(dtype_z),
            },
            "z[i] = y[i] == 0 ? 0 : (x[i] / y[i])",
            "divide",
        )

    def guarded_div(self, a, b):
        from pycuda.gpuarray import _get_common_dtype

        result = a._new_like_me(_get_common_dtype(a, b))

        assert a.shape == b.shape

        func = self.guarded_div_kernel(a.dtype, b.dtype, result.dtype)
        func.prepared_async_call(
            a._grid, a._block, None, a.gpudata, b.gpudata, result.gpudata, a.mem_size
        )

        return result

    def reset(self, rhs, x=None):
        self.rhs = rhs

        if x is None:
            x = np.zeros((self.operator.shape[0],))
        self.x = x

        self.residual = rhs - self.operator(x)

        self.d = self.precon(self.residual)

        # grows at the end
        delta = AsyncInnerProduct(self.residual, self.d, self.pagelocked_allocator)
        self.real_delta_queue = [delta]
        self.delta = delta.gpu_result

    def one_iteration(self, compute_real_residual=False):
        # typed up from J.R. Shewchuk,
        # An Introduction to the Conjugate Gradient Method
        # Without the Agonizing Pain, Edition 1 1/4 [8/1994]
        # Appendix B3

        q = self.operator(self.d)
        myip = gpuarray.dot(self.d, q)
        alpha = self.guarded_div(self.delta, myip)

        self.lc2(1, self.x, alpha, self.d, out=self.x)

        if compute_real_residual:
            self.residual = self.lc2(1, self.rhs, -1, self.operator(self.x))
        else:
            self.lc2(1, self.residual, -alpha, q, out=self.residual)

        s = self.precon(self.residual)
        delta_old = self.delta
        delta = AsyncInnerProduct(self.residual, s, self.pagelocked_allocator)
        self.delta = delta.gpu_result
        beta = self.guarded_div(self.delta, delta_old)

        self.lc2(1, s, beta, self.d, out=self.d)

        if compute_real_residual:
            self.real_delta_queue.append(delta)

    def run(self, max_iterations=None, tol=1e-7, debug_callback=None):
        check_interval = 20

        if max_iterations is None:
            max_iterations = max(3 * check_interval + 1, 10 * self.operator.shape[0])
        real_resid_interval = min(self.operator.shape[0], 50)

        iterations = 0
        delta_0 = None
        while iterations < max_iterations:
            compute_real_residual = iterations % real_resid_interval == 0

            self.one_iteration(compute_real_residual=compute_real_residual)

            if debug_callback is not None:
                what = "it+residual" if compute_real_residual else "it"

                debug_callback(
                    what, iterations, self.x, self.residual, self.d, self.delta
                )

            # do often enough to allow AsyncInnerProduct
            # to progress through (polled) event chain
            rdq = self.real_delta_queue
            if iterations % check_interval == 0:
                if delta_0 is None:
                    delta_0 = rdq[0].get_host_result()
                    if delta_0 is not None:
                        rdq.pop(0)

                if delta_0 is not None:
                    i = 0
                    while i < len(rdq):
                        delta = rdq[i].get_host_result()
                        if delta is not None:
                            if abs(delta) < tol * tol * abs(delta_0):
                                if debug_callback is not None:
                                    debug_callback(
                                        "end",
                                        iterations,
                                        self.x,
                                        self.residual,
                                        self.d,
                                        self.delta,
                                    )
                                return self.x
                            rdq.pop(i)
                        else:
                            i += 1

            iterations += 1

        raise ConvergenceError("cg failed to converge")


def solve_pkt_with_cg(
    pkt_spmv,
    b,
    precon=None,
    x=None,
    tol=1e-7,
    max_iterations=None,
    debug=False,
    pagelocked_allocator=None,
):
    if x is None:
        x = gpuarray.zeros(
            pkt_spmv.shape[0], dtype=pkt_spmv.dtype, allocator=b.allocator
        )
    else:
        x = pkt_spmv.permute(x)

    if pagelocked_allocator is None:
        pagelocked_allocator = drv.pagelocked_empty

    cg = CGStateContainer(pkt_spmv, precon, pagelocked_allocator=pagelocked_allocator)

    cg.reset(pkt_spmv.permute(b), x)

    it_count = [0]
    res_count = [0]

    def debug_callback(what, it_number, x, resid, d, delta):
        if what == "it":
            it_count[0] += 1
        elif what == "it+residual":
            res_count[0] += 1
            it_count[0] += 1

    result = cg.run(max_iterations, tol, debug_callback=debug_callback)

    return pkt_spmv.unpermute(result), it_count[0], res_count[0]



from __future__ import annotations

import numpy as np

from pytools import memoize_method

import pycuda.compiler
import pycuda.driver as drv
import pycuda.gpuarray as array


# {{{ MD5-based random number generation

md5_code = """
/*
 **********************************************************************
 ** Copyright (C) 1990, RSA Data Security, Inc. All rights reserved. **
 **                                                                  **
 ** License to copy and use this software is granted provided that   **
 ** it is identified as the "RSA Data Security, Inc. MD5 Message     **
 ** Digest Algorithm" in all material mentioning or referencing this **
 ** software or this function.                                       **
 **                                                                  **
 ** License is also granted to make and use derivative works         **
 ** provided that such works are identified as "derived from the RSA **
 ** Data Security, Inc. MD5 Message Digest Algorithm" in all         **
 ** material mentioning or referencing the derived work.             **
 **                                                                  **
 ** RSA Data Security, Inc. makes no representations concerning      **
 ** either the merchantability of this software or the suitability   **
 ** of this software for any particular purpose.  It is provided "as **
 ** is" without express or implied warranty of any kind.             **
 **                                                                  **
 ** These notices must be retained in any copies of any part of this **
 ** documentation and/or software.                                   **
 **********************************************************************
 */

/* F, G and H are basic MD5 functions: selection, majority, parity */
#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

/* ROTATE_LEFT rotates x left n bits */
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4 */
/* Rotation is separate from addition to prevent recomputation */
#define FF(a, b, c, d, x, s, ac) \
  {(a) += F ((b), (c), (d)) + (x) + (ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }
#define GG(a, b, c, d, x, s, ac) \
  {(a) += G ((b), (c), (d)) + (x) + (ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }
#define HH(a, b, c, d, x, s, ac) \
  {(a) += H ((b), (c), (d)) + (x) + (ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }
#define II(a, b, c, d, x, s, ac) \
  {(a) += I ((b), (c), (d)) + (x) + (ac); \
   (a) = ROTATE_LEFT ((a), (s)); \
   (a) += (b); \
  }

#define X0 threadIdx.x
#define X1 threadIdx.y
#define X2 threadIdx.z
#define X3 blockIdx.x
#define X4 blockIdx.y
#define X5 blockIdx.z
#define X6 seed
#define X7 i
#define X8 n
#define X9  blockDim.x
#define X10 blockDim.y
#define X11 blockDim.z
#define X12 gridDim.x
#define X13 gridDim.y
#define X14 gridDim.z
#define X15 0

  unsigned int a = 0x67452301;
  unsigned int b = 0xefcdab89;
  unsigned int c = 0x98badcfe;
  unsigned int d = 0x10325476;

  /* Round 1 */
#define S11 7
#define S12 12
#define S13 17
#define S14 22
  FF ( a, b, c, d, X0 , S11, 3614090360); /* 1 */
  FF ( d, a, b, c, X1 , S12, 3905402710); /* 2 */
  FF ( c, d, a, b, X2 , S13,  606105819); /* 3 */
  FF ( b, c, d, a, X3 , S14, 3250441966); /* 4 */
  FF ( a, b, c, d, X4 , S11, 4118548399); /* 5 */
  FF ( d, a, b, c, X5 , S12, 1200080426); /* 6 */
  FF ( c, d, a, b, X6 , S13, 2821735955); /* 7 */
  FF ( b, c, d, a, X7 , S14, 4249261313); /* 8 */
  FF ( a, b, c, d, X8 , S11, 1770035416); /* 9 */
  FF ( d, a, b, c, X9 , S12, 2336552879); /* 10 */
  FF ( c, d, a, b, X10, S13, 4294925233); /* 11 */
  FF ( b, c, d, a, X11, S14, 2304563134); /* 12 */
  FF ( a, b, c, d, X12, S11, 1804603682); /* 13 */
  FF ( d, a, b, c, X13, S12, 4254626195); /* 14 */
  FF ( c, d, a, b, X14, S13, 2792965006); /* 15 */
  FF ( b, c, d, a, X15, S14, 1236535329); /* 16 */

  /* Round 2 */
#define S21 5
#define S22 9
#define S23 14
#define S24 20
  GG ( a, b, c, d, X1 , S21, 4129170786); /* 17 */
  GG ( d, a, b, c, X6 , S22, 3225465664); /* 18 */
  GG ( c, d, a, b, X11, S23,  643717713); /* 19 */
  GG ( b, c, d, a, X0 , S24, 3921069994); /* 20 */
  GG ( a, b, c, d, X5 , S21, 3593408605); /* 21 */
  GG ( d, a, b, c, X10, S22,   38016083); /* 22 */
  GG ( c, d, a, b, X15, S23, 3634488961); /* 23 */
  GG ( b, c, d, a, X4 , S24, 3889429448); /* 24 */
  GG ( a, b, c, d, X9 , S21,  568446438); /* 25 */
  GG ( d, a, b, c, X14, S22, 3275163606); /* 26 */
  GG ( c, d, a, b, X3 , S23, 4107603335); /* 27 */
  GG ( b, c, d, a, X8 , S24, 1163531501); /* 28 */
  GG ( a, b, c, d, X13, S21, 2850285829); /* 29 */
  GG ( d, a, b, c, X2 , S22, 4243563512); /* 30 */
  GG ( c, d, a, b, X7 , S23, 1735328473); /* 31 */
  GG ( b, c, d, a, X12, S24, 2368359562); /* 32 */

  /* Round 3 */
#define S31 4
#define S32 11
#define S33 16
#define S34 23
  HH ( a, b, c, d, X5 , S31, 4294588738); /* 33 */
  HH ( d, a, b, c, X8 , S32, 2272392833); /* 34 */
  HH ( c, d, a, b, X11, S33, 1839030562); /* 35 */
  HH ( b, c, d, a, X14, S34, 4259657740); /* 36 */
  HH ( a, b, c, d, X1 , S31, 2763975236); /* 37 */
  HH ( d, a, b, c, X4 , S32, 1272893353); /* 38 */
  HH ( c, d, a, b, X7 , S33, 4139469664); /* 39 */
  HH ( b, c, d, a, X10, S34, 3200236656); /* 40 */
  HH ( a, b, c, d, X13, S31,  681279174); /* 41 */
  HH ( d, a, b, c, X0 , S32, 3936430074); /* 42 */
  HH ( c, d, a, b, X3 , S33, 3572445317); /* 43 */
  HH ( b, c, d, a, X6 , S34,   76029189); /* 44 */
  HH ( a, b, c, d, X9 , S31, 3654602809); /* 45 */
  HH ( d, a, b, c, X12, S32, 3873151461); /* 46 */
  HH ( c, d, a, b, X15, S33,  530742520); /* 47 */
  HH ( b, c, d, a, X2 , S34, 3299628645); /* 48 */

  /* Round 4 */
#define S41 6
#define S42 10
#define S43 15
#define S44 21
  II ( a, b, c, d, X0 , S41, 4096336452); /* 49 */
  II ( d, a, b, c, X7 , S42, 1126891415); /* 50 */
  II ( c, d, a, b, X14, S43, 2878612391); /* 51 */
  II ( b, c, d, a, X5 , S44, 4237533241); /* 52 */
  II ( a, b, c, d, X12, S41, 1700485571); /* 53 */
  II ( d, a, b, c, X3 , S42, 2399980690); /* 54 */
  II ( c, d, a, b, X10, S43, 4293915773); /* 55 */
  II ( b, c, d, a, X1 , S44, 2240044497); /* 56 */
  II ( a, b, c, d, X8 , S41, 1873313359); /* 57 */
  II ( d, a, b, c, X15, S42, 4264355552); /* 58 */
  II ( c, d, a, b, X6 , S43, 2734768916); /* 59 */
  II ( b, c, d, a, X13, S44, 1309151649); /* 60 */
  II ( a, b, c, d, X4 , S41, 4149444226); /* 61 */
  II ( d, a, b, c, X11, S42, 3174756917); /* 62 */
  II ( c, d, a, b, X2 , S43,  718787259); /* 63 */
  II ( b, c, d, a, X9 , S44, 3951481745); /* 64 */

  a += 0x67452301;
  b += 0xefcdab89;
  c += 0x98badcfe;
  d += 0x10325476;
"""


def rand(shape, dtype=np.float32, stream=None):
    from pycuda.elementwise import get_elwise_kernel
    from pycuda.gpuarray import GPUArray

    result = GPUArray(shape, dtype)

    if dtype == np.float32:
        func = get_elwise_kernel(
            "float *dest, unsigned int seed",
            md5_code
            + """
            #define POW_2_M32 (1/4294967296.0f)
            dest[i] = a*POW_2_M32;
            if ((i += total_threads) < n)
                dest[i] = b*POW_2_M32;
            if ((i += total_threads) < n)
                dest[i] = c*POW_2_M32;
            if ((i += total_threads) < n)
                dest[i] = d*POW_2_M32;
            """,
            "md5_rng_float",
        )
    elif dtype == np.float64:
        func = get_elwise_kernel(
            "double *dest, unsigned int seed",
            md5_code
            + """
            #define POW_2_M32 (1/4294967296.0)
            #define POW_2_M64 (1/18446744073709551616.)

            dest[i] = a*POW_2_M32 + b*POW_2_M64;

            if ((i += total_threads) < n)
            {
              dest[i] = c*POW_2_M32 + d*POW_2_M64;
            }
            """,
            "md5_rng_float",
        )
    elif dtype in [np.int32, np.uint32]:
        func = get_elwise_kernel(
            "unsigned int *dest, unsigned int seed",
            md5_code
            + """
            dest[i] = a;
            if ((i += total_threads) < n)
                dest[i] = b;
            if ((i += total_threads) < n)
                dest[i] = c;
            if ((i += total_threads) < n)
                dest[i] = d;
            """,
            "md5_rng_int",
        )
    else:
        raise NotImplementedError

    func.prepared_async_call(
        result._grid,
        result._block,
        stream,
        result.gpudata,
        np.random.randint(2 ** 31 - 1),
        result.size,
    )

    return result


# }}}

# {{{ CURAND wrapper

try:
    import pycuda._driver as _curand  # used to be separate module
except ImportError:

    def get_curand_version():
        return None


else:
    get_curand_version = _curand.get_curand_version

if get_curand_version() >= (3, 2, 0):
    direction_vector_set = _curand.direction_vector_set
    _get_direction_vectors = _curand._get_direction_vectors

if get_curand_version() >= (4, 0, 0):
    _get_scramble_constants32 = _curand._get_scramble_constants32
    _get_scramble_constants64 = _curand._get_scramble_constants64

# {{{ Base class

gen_template = """
__global__ void %(name)s(%(state_type)s *s, %(out_type)s *d, const size_t n)
{
  const size_t tidx = blockIdx.x*blockDim.x+threadIdx.x;
  const size_t delta = blockDim.x*gridDim.x;
  for (size_t idx = tidx; idx < n; idx += delta)
    d[idx] = curand%(suffix)s(&s[tidx]);
}
"""

gen_log_template = """
__global__ void %(name)s(%(state_type)s *s, %(out_type)s *d, %(in_type)s mean, %(in_type)s stddev, const size_t n)
{
  const size_t tidx = blockIdx.x*blockDim.x+threadIdx.x;
  const size_t delta = blockDim.x*gridDim.x;
  for (size_t idx = tidx; idx < n; idx += delta)
    d[idx] = curand_log%(suffix)s(&s[tidx], mean, stddev);
}
"""

gen_poisson_template = """
__global__ void %(name)s(%(state_type)s *s, %(out_type)s *d, double lambda, const size_t n)
{
  const size_t tidx = blockIdx.x*blockDim.x+threadIdx.x;
  const size_t delta = blockDim.x*gridDim.x;
  for (size_t idx = tidx; idx < n; idx += delta)
    d[idx] = curand_poisson%(suffix)s(&s[tidx], lambda);
}
"""

gen_poisson_inplace_template = """
__global__ void %(name)s(%(state_type)s *s, %(inout_type)s *d, const size_t n)
{
  const size_t tidx = blockIdx.x*blockDim.x+threadIdx.x;
  const size_t delta = blockDim.x*gridDim.x;
  for (size_t idx = tidx; idx < n; idx += delta)
    d[idx] = (%(inout_type)s)(curand_poisson%(suffix)s(&s[tidx], double(d[idx])));
}
"""


random_source = """
// Uses C++ features (templates); do not surround with extern C
#include <curand_kernel.h>

extern "C"
{

%(generators)s

}
"""

random_skip_ahead32_source = """
extern "C" {
__global__ void skip_ahead(%(state_type)s *s, const size_t n, const unsigned int skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
    skipahead(skip, &s[idx]);
}

__global__ void skip_ahead_array(%(state_type)s *s, const size_t n, const unsigned int *skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
      skipahead(skip[idx], &s[idx]);
}
}
"""

random_skip_ahead64_source = """
extern "C" {
__global__ void skip_ahead(%(state_type)s *s, const size_t n, const unsigned long long skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
    skipahead(skip, &s[idx]);
}

__global__ void skip_ahead_array(%(state_type)s *s, const size_t n, const unsigned long long *skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
      skipahead(skip[idx], &s[idx]);
}
}
"""


class _RandomNumberGeneratorBase:
    """
    Class surrounding CURAND kernels from CUDA 3.2.
    It allows for generating random numbers with uniform
    and normal probability function of various types.
    """

    gen_info = [
        ("uniform_int", "unsigned int", ""),
        ("uniform_long", "unsigned long long", ""),
        ("uniform_float", "float", "_uniform"),
        ("uniform_double", "double", "_uniform_double"),
        ("normal_float", "float", "_normal"),
        ("normal_double", "double", "_normal_double"),
        ("normal_float2", "float2", "_normal2"),
        ("normal_double2", "double2", "_normal2_double"),
    ]

    gen_log_info = [
        ("normal_log_float", "float", "float", "_normal"),
        ("normal_log_double", "double", "double", "_normal_double"),
        ("normal_log_float2", "float", "float2", "_normal2"),
        ("normal_log_double2", "double", "double2", "_normal2_double"),
    ]

    gen_poisson_info = [
        ("poisson_int", "unsigned int", ""),
    ]

    gen_poisson_inplace_info = [
        ("poisson_inplace_float", "float", ""),
        ("poisson_inplace_double", "double", ""),
        ("poisson_inplace_int", "unsigned int", ""),
    ]

    def __init__(
        self,
        state_type,
        vector_type,
        generator_bits,
        additional_source,
        scramble_type=None,
    ):
        if get_curand_version() < (3, 2, 0):
            raise OSError("Need at least CUDA 3.2")

        dev = drv.Context.get_device()

        self.block_count = dev.get_attribute(
            pycuda.driver.device_attribute.MULTIPROCESSOR_COUNT
        )

        from pycuda.characterize import has_double_support

        def do_generate(out_type):
            result = True
            if "double" in out_type:
                result = result and has_double_support()
            if "2" in out_type:
                result = result and self.has_box_muller
            return result

        my_generators = [
            (name, out_type, suffix)
            for name, out_type, suffix in self.gen_info
            if do_generate(out_type)
        ]

        if get_curand_version() >= (4, 0, 0):
            my_log_generators = [
                (name, in_type, out_type, suffix)
                for name, in_type, out_type, suffix in self.gen_log_info
                if do_generate(out_type)
            ]

        if get_curand_version() >= (5, 0, 0):
            my_poisson_generators = [
                (name, out_type, suffix)
                for name, out_type, suffix in self.gen_poisson_info
                if do_generate(out_type)
            ]
            my_poisson_inplace_generators = [
                (name, inout_type, suffix)
                for name, inout_type, suffix in self.gen_poisson_inplace_info
                if do_generate(inout_type)
            ]

        generator_sources = [
            gen_template
            % {
                "name": name,
                "out_type": out_type,
                "suffix": suffix,
                "state_type": state_type,
            }
            for name, out_type, suffix in my_generators
        ]

        if get_curand_version() >= (4, 0, 0):
            generator_sources.extend(
                [
                    gen_log_template
                    % {
                        "name": name,
                        "in_type": in_type,
                        "out_type": out_type,
                        "suffix": suffix,
                        "state_type": state_type,
                    }
                    for name, in_type, out_type, suffix in my_log_generators
                ]
            )

        if get_curand_version() >= (5, 0, 0):
            generator_sources.extend(
                [
                    gen_poisson_template
                    % {
                        "name": name,
                        "out_type": out_type,
                        "suffix": suffix,
                        "state_type": state_type,
                    }
                    for name, out_type, suffix in my_poisson_generators
                ]
            )
            generator_sources.extend(
                [
                    gen_poisson_inplace_template
                    % {
                        "name": name,
                        "inout_type": inout_type,
                        "suffix": suffix,
                        "state_type": state_type,
                    }
                    for name, inout_type, suffix in my_poisson_inplace_generators
                ]
            )

        source = (random_source + additional_source) % {
            "state_type": state_type,
            "vector_type": vector_type,
            "scramble_type": scramble_type,
            "generators": "\n".join(generator_sources),
        }

        # store in instance to let subclass constructors get to it.
        self.module = module = pycuda.compiler.SourceModule(source, no_extern_c=True)

        self.generators = {}
        for name, _out_type, _suffix in my_generators:
            gen_func = module.get_function(name)
            gen_func.prepare("PPn")
            self.generators[name] = gen_func
        if get_curand_version() >= (4, 0, 0):
            for name, in_type, _out_type, _suffix in my_log_generators:
                gen_func = module.get_function(name)
                if in_type == "float":
                    gen_func.prepare("PPffn")
                if in_type == "double":
                    gen_func.prepare("PPddn")
                self.generators[name] = gen_func
        if get_curand_version() >= (5, 0, 0):
            for name, _out_type, _suffix in my_poisson_generators:
                gen_func = module.get_function(name)
                gen_func.prepare("PPdn")
                self.generators[name] = gen_func
            for name, _inout_type, _suffix in my_poisson_inplace_generators:
                gen_func = module.get_function(name)
                gen_func.prepare("PPn")
                self.generators[name] = gen_func

        self.generator_bits = generator_bits
        self._prepare_skipahead()

        self.state_type = state_type
        self._state = None

    def _prepare_skipahead(self):
        self.skip_ahead = self.module.get_function("skip_ahead")
        if self.generator_bits == 32:
            self.skip_ahead.prepare("PnI")
        if self.generator_bits == 64:
            self.skip_ahead.prepare("PnQ")
        self.skip_ahead_array = self.module.get_function("skip_ahead_array")
        self.skip_ahead_array.prepare("PnP")

    def _kernels(self):
        return [*list(self.generators.values()), self.skip_ahead, self.skip_ahead_array]

    @property
    @memoize_method
    def generators_per_block(self):
        return min(kernel.max_threads_per_block for kernel in self._kernels())

    @property
    def state(self):
        if self._state is None:
            from pycuda.characterize import sizeof

            data_type_size = sizeof(self.state_type, "#include <curand_kernel.h>")

            self._state = drv.mem_alloc(
                self.block_count * self.generators_per_block * data_type_size
            )

        return self._state

    def fill_uniform(self, data, stream=None):
        if data.dtype == np.float32:
            func = self.generators["uniform_float"]
        elif data.dtype == np.float64:
            func = self.generators["uniform_double"]
        elif data.dtype in [np.int32, np.uint32]:
            func = self.generators["uniform_int"]
        elif data.dtype in [np.int64, np.uint64] and self.generator_bits >= 64:
            func = self.generators["uniform_long"]
        else:
            raise NotImplementedError

        func.prepared_async_call(
            (self.block_count, 1),
            (self.generators_per_block, 1, 1),
            stream,
            self.state,
            data.gpudata,
            data.size,
        )

    def fill_normal(self, data, stream=None):
        if data.dtype == np.float32:
            func_name = "normal_float"
        elif data.dtype == np.float64:
            func_name = "normal_double"
        else:
            raise NotImplementedError

        data_size = data.size
        if self.has_box_muller and data_size % 2 == 0:
            func_name += "2"
            data_size //= 2

        func = self.generators[func_name]

        func.prepared_async_call(
            (self.block_count, 1),
            (self.generators_per_block, 1, 1),
            stream,
            self.state,
            data.gpudata,
            int(data_size),
        )

    def gen_uniform(self, shape, dtype, stream=None):
        result = array.empty(shape, dtype)
        self.fill_uniform(result, stream)
        return result

    def gen_normal(self, shape, dtype, stream=None):
        result = array.empty(shape, dtype)
        self.fill_normal(result, stream)
        return result

    if get_curand_version() >= (4, 0, 0):

        def fill_log_normal(self, data, mean, stddev, stream=None):
            if data.dtype == np.float32:
                func_name = "normal_log_float"
            elif data.dtype == np.float64:
                func_name = "normal_log_double"
            else:
                raise NotImplementedError

            data_size = data.size
            if self.has_box_muller and data_size % 2 == 0:
                func_name += "2"
                data_size //= 2

            func = self.generators[func_name]

            func.prepared_async_call(
                (self.block_count, 1),
                (self.generators_per_block, 1, 1),
                stream,
                self.state,
                data.gpudata,
                mean,
                stddev,
                int(data_size),
            )

        def gen_log_normal(self, shape, dtype, mean, stddev, stream=None):
            result = array.empty(shape, dtype)
            self.fill_log_normal(result, mean, stddev, stream)
            return result

    if get_curand_version() >= (5, 0, 0):

        def fill_poisson(self, data, lambda_value=None, stream=None):
            if lambda_value is None:
                if data.dtype == np.float32:
                    func_name = "poisson_inplace_float"
                elif data.dtype == np.float64:
                    func_name = "poisson_inplace_double"
                elif data.dtype == np.uint32:
                    func_name = "poisson_inplace_int"
                else:
                    raise NotImplementedError
            else:
                if data.dtype == np.uint32:
                    func_name = "poisson_int"
                else:
                    raise NotImplementedError

            func = self.generators[func_name]

            if lambda_value is None:
                func.prepared_async_call(
                    (self.block_count, 1),
                    (self.generators_per_block, 1, 1),
                    stream,
                    self.state,
                    data.gpudata,
                    data.size,
                )
            else:
                func.prepared_async_call(
                    (self.block_count, 1),
                    (self.generators_per_block, 1, 1),
                    stream,
                    self.state,
                    data.gpudata,
                    lambda_value,
                    data.size,
                )

        def gen_poisson(self, shape, dtype, lambda_value, stream=None):
            result = array.empty(shape, dtype)
            self.fill_poisson(result, lambda_value, stream)
            return result

    def call_skip_ahead(self, i, stream=None):
        self.skip_ahead.prepared_async_call(
            (self.block_count, 1),
            (self.generators_per_block, 1, 1),
            stream,
            self.state,
            self.generators_per_block,
            i,
        )

    def call_skip_ahead_array(self, i, stream=None):
        self.skip_ahead_array.prepared_async_call(
            (self.block_count, 1),
            (self.generators_per_block, 1, 1),
            stream,
            self.state,
            self.generators_per_block,
            i.gpudata,
        )


# }}}

# {{{ XORWOW RNG


class _PseudoRandomNumberGeneratorBase(_RandomNumberGeneratorBase):
    def __init__(
        self,
        seed_getter,
        offset,
        state_type,
        vector_type,
        generator_bits,
        additional_source,
        scramble_type=None,
    ):

        super().__init__(
            state_type, vector_type, generator_bits, additional_source
        )

        generator_count = self.generators_per_block * self.block_count
        if seed_getter is None:
            seed = array.to_gpu(
                np.asarray(
                    np.random.randint(0, (1 << 31) - 1, generator_count), dtype=np.int32
                )
            )
        else:
            seed = seed_getter(generator_count)

        if not (
            isinstance(seed, pycuda.gpuarray.GPUArray)
            and seed.dtype == np.int32
            and seed.size == generator_count
        ):
            raise TypeError("seed must be GPUArray of integers of right length")

        p = self.module.get_function("prepare")
        p.prepare("PnPn")

        from pycuda.characterize import has_stack

        has_stack = has_stack()

        if has_stack:
            prev_stack_size = drv.Context.get_limit(drv.limit.STACK_SIZE)

        try:
            if has_stack:
                drv.Context.set_limit(drv.limit.STACK_SIZE, 1 << 14)  # 16k
            try:
                p.prepared_call(
                    (self.block_count, 1),
                    (self.generators_per_block, 1, 1),
                    self.state,
                    generator_count,
                    seed.gpudata,
                    offset,
                )
            except drv.LaunchError:
                raise ValueError("Initialisation failed. Decrease number of threads.")

        finally:
            if has_stack:
                drv.Context.set_limit(drv.limit.STACK_SIZE, prev_stack_size)

    def _prepare_skipahead(self):
        self.skip_ahead = self.module.get_function("skip_ahead")
        self.skip_ahead.prepare("PnQ")
        self.skip_ahead_array = self.module.get_function("skip_ahead_array")
        self.skip_ahead_array.prepare("PnP")
        self.skip_ahead_sequence = self.module.get_function("skip_ahead_sequence")
        self.skip_ahead_sequence.prepare("PnQ")
        self.skip_ahead_sequence_array = self.module.get_function(
            "skip_ahead_sequence_array"
        )
        self.skip_ahead_sequence_array.prepare("PnP")

    def call_skip_ahead_sequence(self, i, stream=None):
        self.skip_ahead_sequence.prepared_async_call(
            (self.block_count, 1),
            (self.generators_per_block, 1, 1),
            stream,
            self.state,
            self.generators_per_block * self.block_count,
            i,
        )

    def call_skip_ahead_sequence_array(self, i, stream=None):
        self.skip_ahead_sequence_array.prepared_async_call(
            (self.block_count, 1),
            (self.generators_per_block, 1, 1),
            stream,
            self.state,
            self.generators_per_block * self.block_count,
            i.gpudata,
        )

    def _kernels(self):
        return (
            [*_RandomNumberGeneratorBase._kernels(self), self.module.get_function("prepare"), self.module.get_function("skip_ahead_sequence"), self.module.get_function("skip_ahead_sequence_array")]
        )


def seed_getter_uniform(n):
    result = pycuda.gpuarray.empty([n], np.int32)
    import random

    value = random.randint(0, 2 ** 31 - 1)
    return result.fill(value)


def seed_getter_unique(n):
    result = np.random.randint(0, 2 ** 31 - 1, n).astype(np.int32)
    return pycuda.gpuarray.to_gpu(result)


xorwow_random_source = """
extern "C" {
__global__ void prepare(%(state_type)s *s, const size_t n,
    %(vector_type)s *v, const size_t o)
{
  const size_t id = blockIdx.x*blockDim.x+threadIdx.x;
  if (id < n)
    curand_init(v[id], id, o, &s[id]);
}
}
"""

xorwow_skip_ahead_sequence_source = """
extern "C" {
__global__ void skip_ahead_sequence(%(state_type)s *s, const size_t n, const unsigned long long skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
    skipahead_sequence(skip, &s[idx]);
}

__global__ void skip_ahead_sequence_array(%(state_type)s *s, const size_t n, const unsigned long long *skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
      skipahead_sequence(skip[idx], &s[idx]);
}
}
"""

if get_curand_version() >= (3, 2, 0):

    class XORWOWRandomNumberGenerator(_PseudoRandomNumberGeneratorBase):
        has_box_muller = True

        def __init__(self, seed_getter=None, offset=0):
            """
            :arg seed_getter: a function that, given an integer count, will yield an `int32`
              :class:`GPUArray` of seeds.
            """

            super().__init__(
                seed_getter,
                offset,
                "curandStateXORWOW",
                "unsigned int",
                32,
                xorwow_random_source
                + xorwow_skip_ahead_sequence_source
                + random_skip_ahead64_source,
            )


# }}}

# {{{ MRG32k3a RNG

mrg32k3a_random_source = """
extern "C" {
__global__ void prepare(%(state_type)s *s, const size_t n,
    %(vector_type)s *v, const size_t o)
{
  const size_t id = blockIdx.x*blockDim.x+threadIdx.x;
  if (id < n)
    curand_init(v[id], id, o, &s[id]);
}
}
"""

mrg32k3a_skip_ahead_sequence_source = """
extern "C" {
__global__ void skip_ahead_sequence(%(state_type)s *s, const size_t n, const unsigned long long skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
    skipahead_sequence(skip, &s[idx]);
}

__global__ void skip_ahead_sequence_array(%(state_type)s *s, const size_t n, const unsigned long long *skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
      skipahead_sequence(skip[idx], &s[idx]);
}

__global__ void skip_ahead_subsequence(%(state_type)s *s, const size_t n, const unsigned long long skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
    skipahead_subsequence(skip, &s[idx]);
}

__global__ void skip_ahead_subsequence_array(%(state_type)s *s, const size_t n, const unsigned long long *skip)
{
  const size_t idx = blockIdx.x*blockDim.x+threadIdx.x;
  if (idx < n)
      skipahead_subsequence(skip[idx], &s[idx]);
}
}
"""

if get_curand_version() >= (4, 1, 0):

    class MRG32k3aRandomNumberGenerator(_PseudoRandomNumberGeneratorBase):
        has_box_muller = True

        def __init__(self, seed_getter=None, offset=0):
            """
            :arg seed_getter: a function that, given an integer count, will yield an `int32`
              :class:`GPUArray` of seeds.
            """

            super().__init__(
                seed_getter,
                offset,
                "curandStateMRG32k3a",
                "unsigned int",
                32,
                mrg32k3a_random_source
                + mrg32k3a_skip_ahead_sequence_source
                + random_skip_ahead64_source,
            )

        def _prepare_skipahead(self):
            super()._prepare_skipahead()
            self.skip_ahead_subsequence = self.module.get_function(
                "skip_ahead_subsequence"
            )
            self.skip_ahead_subsequence.prepare("PnQ")
            self.skip_ahead_subsequence_array = self.module.get_function(
                "skip_ahead_subsequence_array"
            )
            self.skip_ahead_subsequence_array.prepare("PnP")

        def call_skip_ahead_subsequence(self, i, stream=None):
            self.skip_ahead_subsequence.prepared_async_call(
                (self.block_count, 1),
                (self.generators_per_block, 1, 1),
                stream,
                self.state,
                self.generators_per_block * self.block_count,
                i,
            )

        def call_skip_ahead_subsequence_array(self, i, stream=None):
            self.skip_ahead_subsequence_array.prepared_async_call(
                (self.block_count, 1),
                (self.generators_per_block, 1, 1),
                stream,
                self.state,
                self.generators_per_block * self.block_count,
                i.gpudata,
            )

        def _kernels(self):
            return [*_PseudoRandomNumberGeneratorBase._kernels(self), self.module.get_function("skip_ahead_subsequence"), self.module.get_function("skip_ahead_subsequence_array")]


# }}}

# {{{ Sobol RNG


def generate_direction_vectors(count, direction=None):
    if get_curand_version() >= (4, 0, 0):
        if (
            direction == direction_vector_set.VECTOR_64
            or direction == direction_vector_set.SCRAMBLED_VECTOR_64
        ):
            result = np.empty((count, 64), dtype=np.uint64)
        else:
            result = np.empty((count, 32), dtype=np.uint32)
    else:
        result = np.empty((count, 32), dtype=np.uint32)
    _get_direction_vectors(direction, result, count)
    return pycuda.gpuarray.to_gpu(result)


if get_curand_version() >= (4, 0, 0):

    def generate_scramble_constants32(count):
        result = np.empty((count,), dtype=np.uint32)
        _get_scramble_constants32(result, count)
        return pycuda.gpuarray.to_gpu(result)

    def generate_scramble_constants64(count):
        result = np.empty((count,), dtype=np.uint64)
        _get_scramble_constants64(result, count)
        return pycuda.gpuarray.to_gpu(result)


sobol_random_source = """
extern "C" {
__global__ void prepare(%(state_type)s *s, const size_t n,
    %(vector_type)s *v, const size_t o)
{
  const size_t id = blockIdx.x*blockDim.x+threadIdx.x;
  if (id < n)
    curand_init(v[id], o, &s[id]);
}
}
"""


class _SobolRandomNumberGeneratorBase(_RandomNumberGeneratorBase):
    """
    Class surrounding CURAND kernels from CUDA 3.2.
    It allows for generating quasi-random numbers with uniform
    and normal probability function of type int, float, and double.
    """

    has_box_muller = False

    def __init__(
        self,
        dir_vector,
        dir_vector_dtype,
        dir_vector_size,
        dir_vector_set,
        offset,
        state_type,
        vector_type,
        generator_bits,
        sobol_random_source,
    ):
        super().__init__(
            state_type, vector_type, generator_bits, sobol_random_source
        )

        if dir_vector is None:
            dir_vector = generate_direction_vectors(
                self.block_count * self.generators_per_block, dir_vector_set
            )

        if not (
            isinstance(dir_vector, pycuda.gpuarray.GPUArray)
            and dir_vector.dtype == dir_vector_dtype
            and dir_vector.shape
            == (self.block_count * self.generators_per_block, dir_vector_size)
        ):
            raise TypeError("seed must be GPUArray of integers of right length")

        p = self.module.get_function("prepare")
        p.prepare("PnPn")

        from pycuda.characterize import has_stack

        has_stack = has_stack()

        if has_stack:
            prev_stack_size = drv.Context.get_limit(drv.limit.STACK_SIZE)

        try:
            if has_stack:
                drv.Context.set_limit(drv.limit.STACK_SIZE, 1 << 14)  # 16k
            try:
                p.prepared_call(
                    (self.block_count, 1),
                    (self.generators_per_block, 1, 1),
                    self.state,
                    self.block_count * self.generators_per_block,
                    dir_vector.gpudata,
                    offset,
                )
            except drv.LaunchError:
                raise ValueError("Initialisation failed. Decrease number of threads.")

        finally:
            if has_stack:
                drv.Context.set_limit(drv.limit.STACK_SIZE, prev_stack_size)

    def _kernels(self):
        return [*_RandomNumberGeneratorBase._kernels(self), self.module.get_function("prepare")]


scrambledsobol_random_source = """
extern "C" {
__global__ void prepare( %(state_type)s *s, const size_t n,
    %(vector_type)s *v, %(scramble_type)s *scramble, const size_t o)
{
  const size_t id = blockIdx.x*blockDim.x+threadIdx.x;
  if (id < n)
    curand_init(v[id], scramble[id], o, &s[id]);
}
}
"""


class _ScrambledSobolRandomNumberGeneratorBase(_RandomNumberGeneratorBase):
    """
    Class surrounding CURAND kernels from CUDA 4.0.
    It allows for generating quasi-random numbers with uniform
    and normal probability function of type int, float, and double.
    """

    has_box_muller = False

    def __init__(
        self,
        dir_vector,
        dir_vector_dtype,
        dir_vector_size,
        dir_vector_set,
        scramble_vector,
        scramble_vector_function,
        offset,
        state_type,
        vector_type,
        generator_bits,
        scramble_type,
        sobol_random_source,
    ):
        super().__init__(
            state_type, vector_type, generator_bits, sobol_random_source, scramble_type
        )

        if dir_vector is None:
            dir_vector = generate_direction_vectors(
                self.block_count * self.generators_per_block, dir_vector_set
            )

        if scramble_vector is None:
            scramble_vector = scramble_vector_function(
                self.block_count * self.generators_per_block
            )

        if not (
            isinstance(dir_vector, pycuda.gpuarray.GPUArray)
            and dir_vector.dtype == dir_vector_dtype
            and dir_vector.shape
            == (self.block_count * self.generators_per_block, dir_vector_size)
        ):
            raise TypeError("seed must be GPUArray of integers of right length")

        if not (
            isinstance(scramble_vector, pycuda.gpuarray.GPUArray)
            and scramble_vector.dtype == dir_vector_dtype
            and scramble_vector.shape == (self.block_count * self.generators_per_block,)
        ):
            raise TypeError("scramble must be GPUArray of integers of right length")

        p = self.module.get_function("prepare")
        p.prepare("PnPPn")

        from pycuda.characterize import has_stack

        has_stack = has_stack()

        if has_stack:
            prev_stack_size = drv.Context.get_limit(drv.limit.STACK_SIZE)

        try:
            if has_stack:
                drv.Context.set_limit(drv.limit.STACK_SIZE, 1 << 14)  # 16k
            try:
                p.prepared_call(
                    (self.block_count, 1),
                    (self.generators_per_block, 1, 1),
                    self.state,
                    self.block_count * self.generators_per_block,
                    dir_vector.gpudata,
                    scramble_vector.gpudata,
                    offset,
                )
            except drv.LaunchError:
                raise ValueError("Initialisation failed. Decrease number of threads.")

        finally:
            if has_stack:
                drv.Context.set_limit(drv.limit.STACK_SIZE, prev_stack_size)

    def _kernels(self):
        return [*_RandomNumberGeneratorBase._kernels(self), self.module.get_function("prepare")]


if get_curand_version() >= (3, 2, 0):

    class Sobol32RandomNumberGenerator(_SobolRandomNumberGeneratorBase):
        """
        Class surrounding CURAND kernels from CUDA 3.2.
        It allows for generating quasi-random numbers with uniform
        and normal probability function of type int, float, and double.
        """

        def __init__(self, dir_vector=None, offset=0):
            super().__init__(
                dir_vector,
                np.uint32,
                32,
                direction_vector_set.VECTOR_32,
                offset,
                "curandStateSobol32",
                "curandDirectionVectors32_t",
                32,
                sobol_random_source + random_skip_ahead32_source,
            )


if get_curand_version() >= (4, 0, 0):

    class ScrambledSobol32RandomNumberGenerator(
        _ScrambledSobolRandomNumberGeneratorBase
    ):
        """
        Class surrounding CURAND kernels from CUDA 4.0.
        It allows for generating quasi-random numbers with uniform
        and normal probability function of type int, float, and double.
        """

        def __init__(self, dir_vector=None, scramble_vector=None, offset=0):
            super().__init__(
                dir_vector,
                np.uint32,
                32,
                direction_vector_set.SCRAMBLED_VECTOR_32,
                scramble_vector,
                generate_scramble_constants32,
                offset,
                "curandStateScrambledSobol32",
                "curandDirectionVectors32_t",
                32,
                "unsigned int",
                scrambledsobol_random_source + random_skip_ahead32_source,
            )


if get_curand_version() >= (4, 0, 0):

    class Sobol64RandomNumberGenerator(_SobolRandomNumberGeneratorBase):
        """
        Class surrounding CURAND kernels from CUDA 4.0.
        It allows for generating quasi-random numbers with uniform
        and normal probability function of type int, float, and double.
        """

        def __init__(self, dir_vector=None, offset=0):
            super().__init__(
                dir_vector,
                np.uint64,
                64,
                direction_vector_set.VECTOR_64,
                offset,
                "curandStateSobol64",
                "curandDirectionVectors64_t",
                64,
                sobol_random_source + random_skip_ahead64_source,
            )


if get_curand_version() >= (4, 0, 0):

    class ScrambledSobol64RandomNumberGenerator(
        _ScrambledSobolRandomNumberGeneratorBase
    ):
        """
        Class surrounding CURAND kernels from CUDA 4.0.
        It allows for generating quasi-random numbers with uniform
        and normal probability function of type int, float, and double.
        """

        def __init__(self, dir_vector=None, scramble_vector=None, offset=0):
            super().__init__(
                dir_vector,
                np.uint64,
                64,
                direction_vector_set.SCRAMBLED_VECTOR_64,
                scramble_vector,
                generate_scramble_constants64,
                offset,
                "curandStateScrambledSobol64",
                "curandDirectionVectors64_t",
                64,
                "unsigned long long",
                scrambledsobol_random_source + random_skip_ahead64_source,
            )


# }}}

# }}}


# vim: foldmethod=marker
